import numpy as np
import fatpack

# Curtosy of Fatpack examples!
# The different types refer to different types of welds and hot spots.  See Appendix A
curves_in_air = dict(
    reference = "DNV-RP-C203 - Edition October 2024, Table 2-1 S-N curves in air",
    B1= dict(m1=4.0, loga1=15.117, m2=5, Nd=5e6, loga2=17.146, fl=127.21),
    B2= dict(m1=4.0, loga1=14.885, m2=5, Nd=5e6, loga2=16.856, fl=111.3),
    C = dict(m1=3.5, loga1=13.640, m2=5, Nd=5e6, loga2=16.615, fl= 96.21),
    C1= dict(m1=3.5, loga1=13.473, m2=5, Nd=5e6, loga2=16.377, fl= 86.20),
    C2= dict(m1=3.5, loga1=13.301, m2=5, Nd=5e6, loga2=16.130, fl= 76.97),
    D = dict(m1=3.0, loga1=12.164, m2=5, Nd=1e7, loga2=15.606, fl= 52.63),
    E = dict(m1=3.0, loga1=12.010, m2=5, Nd=1e7, loga2=15.350, fl= 46.78),
    F = dict(m1=3.0, loga1=11.855, m2=5, Nd=1e7, loga2=15.091, fl= 41.52),
    F1= dict(m1=3.0, loga1=11.699, m2=5, Nd=1e7, loga2=14.832, fl= 36.84),
    F3= dict(m1=3.0, loga1=11.546, m2=5, Nd=1e7, loga2=14.576, fl= 32.75),
    G = dict(m1=3.0, loga1=11.398, m2=5, Nd=1e7, loga2=14.330, fl= 29.24),
    W1= dict(m1=3.0, loga1=11.261, m2=5, Nd=1e7, loga2=14.101, fl= 26.32),
    W2= dict(m1=3.0, loga1=11.107, m2=5, Nd=1e7, loga2=13.845, fl= 23.39),
    W3= dict(m1=3.0, loga1=10.970, m2=5, Nd=1e7, loga2=13.617, fl= 21.05),
)

curves_in_seawater_with_cathodic_protection = dict(
    reference = "DNV-RP-C203 - Edition October 2024, Table 2-2 S-N in seawater with cathodic protection",
    B1= dict(m1=4.0, loga1=14.977, m2=5, Nd=1e6, loga2=17.222, fl=127.21),
    B2= dict(m1=4.0, loga1=14.745, m2=5, Nd=1e6, loga2=16.931, fl=111.3),
    C = dict(m1=3.5, loga1=13.431, m2=5, Nd=1e6, loga2=16.615, fl= 96.21),
    C1= dict(m1=3.5, loga1=13.264, m2=5, Nd=1e6, loga2=16.377, fl= 86.20),
    C2= dict(m1=3.5, loga1=13.091, m2=5, Nd=1e6, loga2=16.130, fl= 76.97),
    D = dict(m1=3.0, loga1=11.764, m2=5, Nd=1e6, loga2=15.606, fl= 52.63),
    E = dict(m1=3.0, loga1=11.610, m2=5, Nd=1e6, loga2=15.350, fl= 46.78),
    F = dict(m1=3.0, loga1=11.455, m2=5, Nd=1e6, loga2=15.091, fl= 41.52),
    F1= dict(m1=3.0, loga1=11.299, m2=5, Nd=1e6, loga2=14.832, fl= 36.84),
    F3= dict(m1=3.0, loga1=11.146, m2=5, Nd=1e6, loga2=14.576, fl= 32.75),
    G = dict(m1=3.0, loga1=10.998, m2=5, Nd=1e6, loga2=14.330, fl= 29.24),
    W1= dict(m1=3.0, loga1=10.861, m2=5, Nd=1e6, loga2=14.101, fl= 26.32),
    W2= dict(m1=3.0, loga1=10.707, m2=5, Nd=1e6, loga2=13.845, fl= 23.39),
    W3= dict(m1=3.0, loga1=10.570, m2=5, Nd=1e6, loga2=13.617, fl= 21.05),
)

curves_in_seawater_for_free_corrosion = dict(
    reference = "DNV-RP-C203 - Edition October 2024, Table 2-4 S-N in seawater with free corrosion",
    B1= dict(m=3.0, loga=12.436),
    B2= dict(m=3.0, loga=12.262),
    C = dict(m=3.0, loga=12.115),
    C1= dict(m=3.0, loga=11.972),
    C2= dict(m=3.0, loga=11.824),
    D = dict(m=3.0, loga=11.687),
    E = dict(m=3.0, loga=11.533),
    F = dict(m=3.0, loga=11.378),
    F1= dict(m=3.0, loga=11.222),
    F3= dict(m=3.0, loga=11.068),
    G = dict(m=3.0, loga=10.921),
    W1= dict(m=3.0, loga=10.784),
    W2= dict(m=3.0, loga=10.630),
    W3= dict(m=3.0, loga=10.493),
)

def dnv_in_air(name):
    """Returns a DNV endurance curve (SN curve)

    This method returns an endurance curve in air according to 
    table 2-1 in DNV RP-C203.

    Arguments
    ---------
    name : str
        Name of the endurance curve.

    Returns
    -------
    fatpack.BiLinearEnduranceCurve
        Endurance curve corresponding to `name` in DNV RP-C203

    Example
    -------
    >>>curve = DNV_EnduranceCurve.in_air("D")
    >>>N = curve.get_endurance(90.0)

    """

    data = curves_in_air[name]
    curve = fatpack.BiLinearEnduranceCurve(1e6) # 1e6 for MPa to Pa
    curve.Nc = 10 ** data["loga1"] 
    curve.Nd = data["Nd"]
    curve.m1 = data["m1"]
    curve.m2 = data["m2"]
    curve.reference = curves_in_air["reference"]
    return curve
    
    
def dnv_in_seawater_cathodic(name):
    """Returns a DNV endurance curve (SN curve)

    This method returns an endurance curve in seawater with 
    cathodic protection according to table 2-2 in DNV RP-C203.

    Arguments
    ---------
    name : str
        Name of the endurance curve.

    Returns
    -------
    fatpack.BiLinearEnduranceCurve
        Endurance curve corresponding to `name` in DNV RP-C203

    Example
    -------
    >>>curve = DNV_EnduranceCurve.in_seawater_with_cathodic_protection("D")
    >>>N = curve.get_endurance(90.0)

    """
    data = curves_in_seawater_with_cathodic_protection[name]
    curve = fatpack.BiLinearEnduranceCurve(1e6) # 1e6 for MPa to Pa
    curve.Nc = 10 ** data["loga1"]
    curve.Nd = data["Nd"]
    curve.m1 = data["m1"]
    curve.m2 = data["m2"]
    ref = curves_in_seawater_with_cathodic_protection["reference"]
    curve.reference = ref
    return curve

def dnv_in_seawater(name):
    """Returns a DNV endurance curve (SN curve)

    This method returns an endurance curve in seawater for 
    free corrosion according to table 2-4 in DNV RP-C203.

    Arguments
    ---------
    name : str
        Name of the endurance curve.

    Returns
    -------
    fatpack.LinearEnduranceCurve
        Endurance curve corresponding to `name` in DNV RP-C203

    Example
    -------
    >>>curve = DNV_EnduranceCurve.in_seawater_for_free_corrosion("D")
    >>>N = curve.get_endurance(90.0)

    """
    data = curves_in_seawater_for_free_corrosion[name]
    curve = fatpack.LinearEnduranceCurve(1e6) # 1e6 for MPa to Pa
    curve.Nc = 10 ** data["loga"]
    curve.m = data["m"]
    ref = curves_in_seawater_for_free_corrosion["reference"]
    curve.reference = ref
    return curve

# Could use a dict or namedtuple here, but this standardizes things a bit better for users
class FatigueParams:
    """Data structure of parameters needed by fatigue calculation and standard calculations."""

    def __init__(self, **kwargs):
        """
        Creates an instance of `FatigueParams`.  There are three ways to initialize a FatigueParams
        instance with an associated S-N curve in the `fatpack` library:
        
        1. Specify a curve from DNV-RP-C203, Fatigue in Offshore Steel Structures.
           Input keywords are "dnv_type" = (one of) ['air', 'seawater', 'cathodic'] and
           "dnv_name" = (one of) [B1, B2, C, C1, C2, D, E, G F1, F3, G, W1, W2, W3]

        2. Specify the slope of the S-N curve and a point on the curve.
           Required keywords are "slope", "Nc" and "Sc".  Assumes a linear S-N curve.

        3. Specify the slope and the S-intercept point assuming a perflectly linear S-N curve
           (which might not be the actual ultimate failure stress of the material.
           Required keywords are "slope" and "S_intercept".
        
        Parameters
        ----------
        load2stress : float (optional)
            Linear scaling coefficient to convert an applied load to stress such that S = load2stress * L.
            Required for Goodman correction calculations of DELs and for all Damage calculations
        dnv_type : string (optional)
            Type of S-N curve to use 'air' for structures in air, 'seawater' for structures in seawater, or
            'cathodic' for structures with cathodic protection in seawater.  Must also specify "dnv_name"
            From DNV-RP-C203, Fatigue of Metal Structures, - Edition October 2024
        dnv_name : string (optional)
            Grade of metal and hot spot exposure to use: [B1, B2, C, C1, C2, D, E, G F1, F3, G, W1, W2, W3].  Must also specify "dnv_type"
            From DNV-RP-C203, Fatigue of Metal Structures, - Edition October 2024
        slope : float (optional)
            Wohler exponent in the traditional SN-curve of S = A * N ^ -(1/m).  Must either specify Sc-Nc or S_intercept.
        Sc : float (optional)
            The S-value (amplitude of stress osciallation) on a point on the linear S-N curve.  Must also specify "Nc" and "slope"
        Nc : float (optional)
            The N-value (number of endurance cycles to failure) on a point on the linear S-N curve.  Must also specify "Sc" and "slope"
        ultimate_stress : float (optional)
            Ultimate stress for use in Goodman equivalent stress calculation
        S_intercept : float (optional)
            Stress-axis intercept of log-log S-N Wohler curve. Taken as ultimate stress unless specified
        rainflow_bins : int (optional)
            Number of bins used in rainflow analysis.
            Default: 100
        goodman_correction: boolean (optional)
            Whether to apply Goodman mean correction to loads and stress
            Default: False
        """

        self.load2stress   = kwargs.get("load2stress", 1.0)
        dnv_name           = kwargs.get("dnv_name", "").upper()
        dnv_type           = kwargs.get("dnv_type", "").lower()
        slope              = np.abs( kwargs.get("slope", 4.0) )
        Sc                 = kwargs.get("Sc", None)
        Nc                 = kwargs.get("Nc", None)
        self.S_ult         = kwargs.get("ultimate_stress", 1.0)
        temp               = kwargs.get("S_intercept", 0.0)
        S_intercept        = temp if temp > 0.0 else self.S_ult
        self.bins          = kwargs.get("rainflow_bins", 100)
        self.bins          = kwargs.get("bins", self.bins)
        self.goodman       = kwargs.get("goodman_correction", False)
        self.goodman       = kwargs.get("goodman", self.goodman)

        for k in kwargs:
            if k not in ["load2stress", "dnv_name", "dnv_type",
                         "slope", "Sc", "Nc", "ultimate_stress",
                         "S_intercept", "rainflow_bins", "bins",
                         "goodman_correction", "goodman"]:
                print(f"Unknown keyword argument, {k}")

        if dnv_name is not None and len(dnv_name) > 0:
            if dnv_type.find("cath") >= 0:
                self.curve = dnv_in_seawater_cathodic(dnv_name)
            elif dnv_type.find("sea") >= 0:
                self.curve = dnv_in_seawater(dnv_name)
            elif dnv_type.find("air") >= 0:
                self.curve = dnv_in_air(dnv_name)
            else:
                raise ValueError(f'Unknown DNV RP-C203 curve type, {dnv_type}. Expected [air, seawater, or cathodic]')

        elif Sc is not None and Sc > 0.0 and Nc is not None and Nc > 0.0:
            # Set fatpack linear curve from S-N point
            self.curve = fatpack.LinearEnduranceCurve(Sc)
            self.curve.m = slope
            self.curve.Nc = Nc

        else:
            # Set fatpath curve from S-intercept
            self.curve = fatpack.LinearEnduranceCurve(S_intercept)
            self.curve.m = slope
            self.curve.Nc = 1

        
    def copy(self):
        newobj = FatigueParams(load2stress=self.load2stress,
                               ultimate_stress=self.S_ult,
                               bins=self.bins,
                               goodman=self.goodman)
        newobj.curve = self.curve
        return newobj

    
    def get_rainflow_matrix(self, chan, bins=20):
        """
        Computes ranflow counts and bins for input `chan`.
        Typically used internally to the class, but can be used for plotting and debugging too
        
        Parameters
        ----------
        chan : np.array
            Channel time series to calculate DEL for.
        bins : int
            Number of bins used in rainflow analysis.
        goodman: boolean (optional)
            Whether to apply Goodman mean correction to loads and stress
            Default: False
        S_ult: float (optional)
            Ultimate stress/load for the material
        """

        S, Mrf = fatpack.find_rainflow_ranges(chan, k=256, return_means=True)
        data_arr = np.c_[S, Mrf]
        rowbin, colbin, rfcmat = fatpack.find_rainflow_matrix(data_arr, bins, bins, return_bins=True)

        X, Y = np.meshgrid(rowbin, colbin, indexing='ij')

        return rfcmat, X, Y

    
    def get_rainflow_counts(self, chan, bins, S_ult=None, goodman=False):
        """
        Computes ranflow counts and bins for input `chan`.
        Typically used internally to the class, but can be used for plotting and debugging too
        
        Parameters
        ----------
        chan : np.array
            Channel time series to calculate DEL for.
        bins : int
            Number of bins used in rainflow analysis.
        goodman: boolean (optional)
            Whether to apply Goodman mean correction to loads and stress
            Default: False
        S_ult: float (optional)
            Ultimate stress/load for the material

        Returns
        -------
        N, S : 1darray
        The count and the characteristic value for the ranges.
        """

        try:
            ranges, Mrf = fatpack.find_rainflow_ranges(chan, k=256, return_means=True)
        except Exception:
            ranges = Mrf = np.zeros(1)
            
        if goodman:
            if S_ult is None:
                S_ult = self.S_ult
                
            if S_ult == 0.0:
                raise ValueError('Must specify an ultimate_stress to use Goodman correction')
                
            ranges = fatpack.find_goodman_equivalent_stress(ranges, Mrf, S_ult)

        success = False
        while not success:
            try:
                N, S = fatpack.find_range_count(ranges, bins)
                success = True
            except ValueError:
                bins *= 0.5
                if bins < 1:
                    print(ranges)
                    print(bins)
                    raise Exception("Failed to find bins for ranges")
                else:
                    bins = int(bins)
                    
        return N, S

    
    def compute_del(self, chan, elapsed_time, **kwargs):
        """
        Computes damage equivalent load of input `chan`.

        Parameters
        ----------
        chan : np.array
            Channel time series to calculate DEL for.
        rainflow_bins : int (optional kwargs)
            Number of bins used in rainflow analysis.
            Default: 100
        goodman_correction: boolean (optional kwargs)
            Whether to apply Goodman mean correction to loads and stress
            Default: False
        """

        bins          = kwargs.get("rainflow_bins", self.bins)
        bins          = kwargs.get("bins", bins)
        goodman       = kwargs.get("goodman_correction", self.goodman)
        goodman       = kwargs.get("goodman", goodman)

        for k in kwargs:
            if k not in ["rainflow_bins", "bins", "goodman_correction", "goodman"]:
                print(f"Unknown keyword argument, {k}")
            
        # Default return values
        DEL = np.nan
        D   = np.nan

        if np.all(np.isnan(chan)):
            return DEL, D

        # Working with loads for DELs
        load2 = 1.0 if self.load2stress==0.0 else np.abs(self.load2stress)
        Nrf, Frf = self.get_rainflow_counts(chan, bins, S_ult=self.S_ult/load2, goodman=goodman)

        slope = self.curve.m1 if hasattr(self.curve, 'm1') else self.curve.m
        DELs = Frf ** slope * Nrf / elapsed_time
        DEL = DELs.sum() ** (1.0 / slope)
        # With fatpack do:
        #curve = fatpack.LinearEnduranceCurve(1.)
        #curve.m = slope
        #curve.Nc = elapsed
        #DEL = curve.find_miner_sum(np.c_[Frf, Nrf]) ** (1 / slope)
            
        return DEL
    
        
    def compute_damage(self, chan, **kwargs):
        """
        Computes Palmgren-Miner damage of input `chan`.

        Parameters
        ----------
        chan : np.array
            Channel time series to calculate DEL for.
        rainflow_bins : int (optional kwargs)
            Number of bins used in rainflow analysis.
            Default: 100
        goodman_correction: boolean (optional kwargs)
            Whether to apply Goodman mean correction to loads and stress
            Default: False
        """

        bins          = kwargs.get("rainflow_bins", self.bins)
        bins          = kwargs.get("bins", bins)
        goodman       = kwargs.get("goodman_correction", self.goodman)
        goodman       = kwargs.get("goodman", goodman)

        for k in kwargs:
            if k not in ["rainflow_bins", "bins", "goodman_correction", "goodman"]:
                print(f"Unknown keyword argument, {k}")
            
        # Default return values
        D   = np.nan

        if np.all(np.isnan(chan)):
            return D

        if np.abs(self.load2stress) == 0.0:
            raise ValueError('Must specify load2stress value to compute Damage')

        Nrf, Srf = self.get_rainflow_counts(chan*self.load2stress, bins, goodman=goodman)

        D = self.curve.find_miner_sum(np.c_[Srf, Nrf])
        
        return D

    
    def get_stress(self, NN):
        return self.curve.get_stress(NN)





import os

import pandas as pd
import ruamel.yaml as ry
from scipy.special import gamma
from scipy.stats import rayleigh, weibull_min
import numpy as np


def weibull_mean(x, k, xbar, kind="pdf"):
    """
    Weibull probability/cumulative density function using desired mean value.
    
    Parameters
    ----------
    x : list or np.array
        Vector of points to evaluate CDF
    k : float
        Weibull shape factor
    xbar : float
        Weibull mean value
    """
    scale = xbar / gamma(1.0 + 1.0 / k)
    if kind.lower() == "pdf":
        return weibull_min.pdf(x, k, loc=0, scale=scale)
    elif kind.lower() == "cdf":
        return weibull_min.cdf(x, k, loc=0, scale=scale)
    else:
        raise ValueError("Expected pdf/cdf for kind")

def rayleigh_mean(x, xbar, kind="pdf"):
    """
    Weibull probability/cumulative density function using desired mean value.
    
    Parameters
    ----------
    x : list or np.array
        Vector of points to evaluate CDF
    xbar : float
        Rayleigh mean value
    """
    scale = np.sqrt(2.0/np.pi) * xbar
    if kind.lower() == "pdf":
        return rayleigh.pdf(x, loc=0, scale=scale)
    elif kind.lower() == "cdf":
        return rayleigh.cdf(x, loc=0, scale=scale)
    else:
        raise ValueError("Expected pdf/cdf for kind")

def df2dict(df):
    """
    Build dictionary from pandas dataframe
    Parameters:
    -----------
    df: DataFrame
        DataFrame with summary stats (probably). Cannot have len(multiindex) > 3
    Returns:
    --------
    dfd: dict
        dictionary containing re-structured dataframe inputs
    """

    if len(df.columns[0]) == 3:
        dfd = [
            {
                level: df.xs(dset, axis=1).xs(level, axis=1).to_dict("list")
                for level in df.columns.levels[1]
            }
            for dset in df.columns.levels[0]
        ]

    elif len(df.columns[0]) == 2:
        dfd = {
            level: df.xs(level, axis=1).to_dict("list")
            for level in df.columns.levels[0]
        }

    elif len(df.columns[0]) == 1:
        dfd = df.to_dict("list")

    else:
        raise TypeError(
            "Converting DataFrames with multiindex > 3 to dictionaries is not supported"
        )

    dfd["meta"] = {"filename": list(df.index)}

    return dfd


def dict2df(sumstats, names=None):
    """
    Build pandas datafrom from list of summary statistics.

    Inputs:
    -------
    sumstats: list
        List of the dictionaries loaded from post_process.load_yaml
    names: list, optional
        List of names for each run. len(sumstats)=len(names)

    Returns:
    --------
    df: pd.DataFrame
        pandas dataframe
    """

    if isinstance(sumstats, list):
        if not names:
            names = ["dataset_" + str(i) for i in range(len(sumstats))]
        data_dict = {
            (name, outerKey, innerKey): values
            for name, sumdata in zip(names, sumstats)
            for outerKey, innerDict in sumdata.items()
            for innerKey, values in innerDict.items()
        }

    else:
        data_dict = {
            (outerKey, innerKey): values
            for outerKey, innerDict in sumstats.items()
            for innerKey, values in innerDict.items()
        }

    # Make dataframe
    df = pd.DataFrame(data_dict)

    return df

def save_yaml(outdir, fname, data):
    """
    Save data as '.yaml' file.

    Parameters:
    -----------
    outdir : str
        Directory to save data.
    fname : str
        Filename for output yaml.
    data : dict
        Data to save to yaml.
    """

    filepath = os.path.join(outdir, fname)
    if not os.path.isdir(outdir):
        os.makedirs(outdir)

    #with open(filepath, "w") as f:
    #    yaml.dump(data, f, Dumper=Dumper, default_flow_style=False)
        
    # Write yaml with updated values
    myyaml = ry.YAML()
    myyaml.default_flow_style = None
    myyaml.width = float("inf")
    myyaml.indent(mapping=4, sequence=6, offset=3)
    myyaml.allow_unicode = False
    with open(filepath, "w", encoding="utf-8") as f:
        myyaml.dump(data, f)


def load_yaml(filepath):
    """
    Import a .yaml file.

    Parameters:
    -----------
    filepath : str
        File to load.
    """

    reader = ry.YAML(typ="safe", pure=True)
    with open(filepath, "r", encoding="utf-8") as f:
        data = reader.load(f)
    
    #with open(filepath) as f:
    #    data = yaml.load(f, Loader=Loader)

    return data



def yaml2df(filename, names=[]):
    """
    Read yaml containing summary stats into dataframe
    Parameters:
    -------
    filename:
        Name of yaml file to load.
    """

    data_dict = load_yaml("test.yaml", package=0)

    level = data_dict
    li = 0  # level index
    while isinstance(level, dict):
        li = li + 1
        if isinstance(level, dict):
            level = level[list(level.keys())[0]]

    data_list = []
    if li == 3:
        for key in data_dict.keys():
            data_list.append(data_dict[key])
        if not names:
            names = ["dataset_" + str(i) for i in range(len(data_list))]

    elif li == 2:
        data_list.append(data_dict)
        names = []
    else:
        raise TypeError(
            "{} is improperly structured for yaml2df.".format(filename)
        )

    df = dict2df(data_list, names=names)

    return df


def get_windspeeds(case_matrix, return_df=False):
    """
    Find windspeeds from case matrix
    Parameters:
    ----------
    case_matrix: dict
        case matrix data loaded from wisdem.aeroelasticse.Util.FileTools.load_yaml

    Returns:
    --------
    windspeed: list
        list of wind speeds
    seed: seed
        list of wind seeds
    IECtype: list
        list of IEC types
    case_matrix: pd.DataFrame
        case matrix dataframe with appended wind info
    """

    if isinstance(case_matrix, dict):
        cmatrix = case_matrix

    elif isinstance(case_matrix, pd.DataFrame):
        cmatrix = case_matrix.to_dict("list")

    else:
        raise TypeError("case_matrix must be a dict or pd.DataFrame.")

    # loop through and parse each inflow filename text entry to get wind and seed #
    if ("InflowWind", "HWindSpeed") in cmatrix:
        windspeed = [float(m) for m in cmatrix[("InflowWind", "HWindSpeed")]]
        seed = [float(m) for m in cmatrix[("TurbSim", "RandSeed1")]]
        IECtype = [None] * len(seed)

    elif ("InflowWind", "Filename_Uni") in cmatrix:
        windspeed = []
        seed = []
        IECtype = []
        for fname in cmatrix[("InflowWind", "Filename_Uni")]:
            if ".bts" in fname:
                obj = fname.split("U")[-1].split("_")
                obj2 = obj[1].split("Seed")[-1].split(".bts")
                windspeed.append(float(obj[0]))
                seed.append(float(obj2[0]))
                if "NTM" in fname:
                    IECtype.append("NTM")
                elif "ETM" in fname:
                    IECtype.append("NTM")
                elif "EWM" in fname:
                    IECtype.append("EWM")
                else:
                    IECtype.append("")

            elif "ECD" in fname:
                obj = fname.split("U")[-1].split("_D")[0].split(".wnd")[0]
                windspeed.append(float(obj))
                seed.append([])
                IECtype.append("ECD")

            elif "EWS" in fname:
                obj = fname.split("U")[-1].split("_D")[0].split(".wnd")[0]
                windspeed.append(float(obj))
                seed.append([])
                IECtype.append("EWS")

            else:
                print("Shouldn't get here")
                print(fname)
                breakpoint()

    if return_df:
        case_matrix = pd.DataFrame(case_matrix)
        case_matrix[("InflowWind", "WindSpeed")] = windspeed
        case_matrix[("InflowWind", "Seed")] = seed
        case_matrix[("InflowWind", "IECtype")] = IECtype

        return windspeed, seed, IECtype, case_matrix

    else:
        return windspeed, seed, IECtype


def convert_summary_stats(df):

    out = {}
    melted = df.to_dict(orient="list")
    for col, stat in melted:

        col = str(col)
        stat = str(stat)
        try:
            out[col]

        except KeyError:
            out[col] = {}

        out[col][stat] = list(melted[(col, stat)])
        out["meta"] = {"filename": list(df.index)}

    return out


# class wsPlotting(object):
#     '''
#     General plotting scripts.
#     '''

#     def __init__(self):
#         pass

#     def stat_curve(self, windspeeds, stats, plotvar, plottype, stat_idx=0, names=[]):
#         '''
#         Plot the turbulent power curve for a set of data.
#         Can be plotted as bar (good for comparing multiple cases) or line

#         Parameters:
#         -------
#         windspeeds: list-like
#             List of wind speeds to plot
#         stats: list, dict, or pd.DataFrame
#             Dict (single case), list(multiple cases), df(single or multiple cases) containing
#             summary statistics.
#         plotvar: str
#             Type of variable to plot
#         plottype: str
#             bar or line
#         stat_idx: int, optional
#             Index of datasets in stats to plot from

#         Returns:
#         --------
#         fig: figure handle
#         ax: axes handle
#         '''

#         # Check for valid inputs
#         if isinstance(stats, dict):
#             stats_df = pdTools.dict2df(stats)
#             if any((stat_inds > 0) or (isinstance(stat_inds, list))):
#                 print('WARNING: stat_ind = {} is invalid for a single stats dictionary. Defaulting to stat_inds=0.')
#                 stat_inds = 0
#         elif isinstance(stats, list):
#             stats_df = pdTools.dict2df(stats)
#         elif isinstance(stats, pd.DataFrame):
#             stats_df = stats
#         else:
#             raise TypeError(
#                 'Input stats must be a dictionary, list, or pd.DataFrame containing OpenFAST output statistics.')


#         # Check windspeed length
#         if len(windspeeds) == len(stats_df):
#             ws = windspeeds
#         elif int(len(windspeeds)/len(stats_df.columns.levels[0])) == len(stats_df):
#             ws = windspeeds[0:len(stats_df)]
#         else:
#             raise ValueError('Length of windspeeds is not the correct length for the input statistics')

#         # Get statistical data for desired plot variable
#         if plotvar in stats_df.columns.levels[0]:
#             sdf = stats_df.loc[:, (plotvar, slice(None))].droplevel([0], axis=1)
#         elif plotvar in stats_df.columns.levels[1]:
#             sdf = stats_df.loc[:, (slice(None), plotvar, slice(None))].droplevel([1], axis=1)
#         else:
#             raise ValueError("{} does not exist in the input statistics.".format(plotvar))

#         # Add windspeeds to data
#         sdf['WindSpeeds']= ws
#         # Group by windspeed and average each statistic (for multiple seeds)
#         sdf = sdf.groupby('WindSpeeds').mean()
#         # Final wind speed values
#         pl_windspeeds=sdf.index.values

#         if plottype == 'bar':
#             # Define mean and std dataframes
#             means = sdf.loc[:, (slice(None), 'mean')].droplevel(1, axis=1)
#             std = sdf.loc[:, (slice(None), 'std')].droplevel(1, axis=1)
#             # Plot bar charts
#             fig, ax = plt.subplots()
#             means.plot.bar(yerr=std, ax=ax, title=plotvar, capsize=2)
#             ax.legend(names,loc='upper left')

#         if plottype == 'line':
#             # Define mean, min, max, and std dataframes
#             means = sdf.loc[:, (sdf.columns.levels[0][stat_idx], 'mean')]
#             smax = sdf.loc[:, (sdf.columns.levels[0][stat_idx], 'max')]
#             smin = sdf.loc[:, (sdf.columns.levels[0][stat_idx], 'min')]
#             std = sdf.loc[:, (sdf.columns.levels[0][stat_idx], 'std')]

#             fig, ax = plt.subplots()
#             ax.errorbar(pl_windspeeds, means, [means - smin, smax - means],
#                          fmt='k', ecolor='gray', lw=1, capsize=2)
#             means.plot(yerr=std, ax=ax,
#                         capsize=2, lw=3,
#                         elinewidth=2,
#                         title=names[0] + ' - ' + plotvar)
#             plt.grid(lw=0.5, linestyle='--')

#         return fig, ax


#     def distribution(self, fast_data, channels, caseid, names=None, kde=True):
#         '''
#         Distributions of data from desired fast runs and channels

#         Parameters
#         ----------
#         fast_data: dict, list
#             List or Dictionary containing OpenFAST output data from desired cases to compare
#         channels: list
#             List of strings of OpenFAST output channels e.g. ['RotSpeed','GenTq']
#         caseid: list
#             List of caseid's to compare
#         names: list, optional
#             Names of the runs to compare
#         fignum: ind, (optional)
#             Specified figure number. Useful to plot over previously made plot

#         Returns:
#         --------
#         fig: figure handle
#         ax: axes handle
#         '''
#         # Make sure input types allign
#         if isinstance(fast_data, dict):
#             fd = [fast_data]
#         elif isinstance(fast_data, list):
#             if len(caseid) == 1:
#                 fd = [fast_data[caseid[0]]]
#             else:
#                 fd = [fast_data[case] for case in caseid]
#         else:
#             raise ValueError('fast_data is an improper data type')


#         # if not names:
#         #     names = [[]]*len(fd)

#         for channel in channels:
#             fig, ax = plt.subplots()
#             for idx, data in enumerate(fd):
#                 # sns.kdeplot(data[channel], shade=True, label='case '+ str(idx))
#                 sns.distplot(data[channel], kde=kde, label='case ' + str(idx))  # For a histogram
#                 ax.set_title(channel + ' distribution')

#                 units = data['meta']['attribute_units'][data['meta']['channels'].index(channel)]
#                 ax.set_xlabel('{} [{}]'.format(channel, units))
#                 ax.grid(True)
#             if names:
#                 ax.legend(names)

#         return fig, ax


# def plot_load_ranking(self, load_rankings, case_matrix, classifier_type,
#                     classifier_names=[], n_rankings=10, caseidx_labels=False):
#     '''
#     Plot load rankings

#     Parameters:
#     -----------
#     load_rankings: list, dict, or pd.DataFrame
#         Dict (single case), list(multiple cases), df(single or multiple cases) containing
#         load rankings.
#     case_matrix: dict or pdDataFrame
#         Information mapping classifiers to load_rankings.
#         NOTE: the case matrix must have wind speeds in ('InflowWind','WindSpeeds') if you
#         wish to plot w.r.t. wind speed
#     classifier_type: tuple or str
#         classifier to denote load ranking cases. e.g. classifier_type=('IEC','DLC') will separate
#         the load rankings by DLC type, assuming the case matrix is properly set up to map each
#         DLC to the load ranking case
#     classifier_names: list, optional
#         Naming conventions for each classifier type for plotting purposes
#     n_rankings: int, optional
#         number of load rankings to plot
#     caseidx_labels: bool, optional
#         label x-axis with case index if True. If false, will try to plot with wind speed labels
#         if they exist, then fall abck to case indeces.

#     TODO: Save figs
#     '''

#     # flag_DLC_name = False
#     # n_rankings = 10
#     # fig_ext = '.pdf'
#     # font_size = 10
#     # classifier_type = ('ServoDyn', 'DLL_FileName')
#     # classifiers = list(set(cmw[classifier_type]))
#     # classifier_names = ['ROSCO', 'legacy']

#     # Check for valid inputs
#     if isinstance(load_rankings, dict):
#         load_ranking_df = pdTools.dict2df(load_rankings)
#     elif isinstance(load_rankings, list):
#         load_ranking_df = pdTools.dict2df(load_rankings)
#     elif isinstance(load_rankings, pd.DataFrame):
#         load_ranking_df = load_rankings
#     else:
#         raise TypeError(
#             'Input stats must be a dictionary, list, or pd.DataFrame containing OpenFAST output statistics.')

#     # Check multiindex size
#     if len(load_ranking_df) == 2:
#         load_ranking_df = pd.concat([load_ranking_df], keys=[dataset_0])

#     # Check for classifier_names
#     classifiers = list(set(case_matrix[classifier_type]))
#     if not classifier_names:
#         classifier_names = ['datatset_{}'.format(idx) for idx in range(len(classifiers))]

#     # Check for wind speeds in case_matrix
#     if not caseidx_labels:
#         try:
#             windspeeds = case_matrix[('InflowWind','WindSpeed')]
#         except:
#             print('Unable to find wind speeds in case_matrix, plotting w.r.t case index')
#             caseidx_labels=True


#     # Define a color map
#     clrs = np.array([[127, 60, 141],
#                     [17, 165, 121],
#                     [57, 105, 172],
#                     [242, 183, 1],
#                     [231, 63, 116],
#                     [128, 186, 90],
#                     [230, 131, 16],
#                     [256, 256, 256]]) / 256.

#     # Get channel names
#     channels = load_ranking_df.columns.levels[1]

#     # initialize some variables
#     colors = np.zeros((n_rankings, 3))
#     labels = [''] * n_rankings
#     labels_index = [''] * n_rankings
#     fig_list = []
#     ax_list = []
#     # --- Generate plots ---
#     for cidx, channel in enumerate(channels):
#         # Pull out specific channel
#         cdf = load_ranking_df.loc[:, (slice(None), channel, slice(None))].droplevel(1, axis=1)
#         # put the load ranking from each dataset in a list so we can combine them
#         cdf_list = [cdf[dataset] for dataset in cdf.columns.levels[0]]
#         chan_df = pd.concat(cdf_list) # combine all load rankings
#         chan_stats = chan_df.columns.values # pull out the names of the columns
#         chan_df.sort_values(by=chan_stats[0], ascending=False, inplace=True) # sort
#         chan_df.reset_index(inplace=True, drop=True) # re-index

#         # find colors and labels for plots
#         for i in range(n_rankings):
#             classifier = case_matrix[classifier_type][chan_df[chan_stats[1]][i]]
#             colors[i, :] = clrs[min(len(clrs), classifiers.index(classifier))]

#             if not caseidx_labels:
#                 ws = windspeeds[chan_df[chan_stats[1]][i]]
#                 labels[i] = classifier_names[classifiers.index(classifier)] + ' - ' + str(ws) + ' m/s'
#             else:
#                 labels[i] = classifier_names[classifiers.index(classifier)] + ' - Case ' + str(chan_df[chan_stats[1]][i])
#     #         labels_index = ['case {}'.format(case) for case in chan_df[chan_stats[1]][0:n_rankings]]

#         # make plot
#         fig, ax = plt.subplots()
#         chan_df[chan_stats[0]][0:n_rankings].plot.bar(color=colors)
#         ax.set_ylabel(channel)
#         ax.set_xticklabels(labels, rotation=45, ha='right')
#         plt.draw()

#         fig_list.append(fig)
#         ax_list.append(ax)

#     #     if case_idx_labels:
#     #         ax.set_xlabel('DLC [-]', fontsize=font_size+2, fontweight='bold')
#     # #         ax.set_xticklabels(np.arange(n_rankings), labels=labels)
#     #         ax.set_xticklabels(labels)
#     #     else:
#     #         #         ax.set_xticklabels(np.arange(n_rankings), labels=labels)

#     return fig_list, ax_list



import os
import fnmatch
import numpy as np
import pandas as pd
from scipy import stats, signal
import numexpr as ne
import fatpack

def dataproperty(f):
    @property
    def wrapper(self, *args, **kwargs):
        if getattr(self, "data", None) is None:
            raise AttributeError("Output has not been read yet.")
        return f(self, *args, **kwargs)

    return wrapper


class AeroelasticOutput:
    """Base timeseries aeroelastic simulation data output class."""

    def __init__(self, datain=None, channelsin=None, **kwargs):
        """
        Creates an instance of `AeroelasticOutput`.

        Parameters
        ----------
        data : list
            List of OpenFAST output filepaths or dicts of OpenFAST outputs.
        channels : list (optional)
            List of OpenFAST output filepaths or dicts of OpenFAST outputs.
        units : list (optional)
            List of OpenFAST output filepaths or dicts of OpenFAST outputs.
        name : list (optional)
            List of OpenFAST output filepaths or dicts of OpenFAST outputs.
        description : list (optional)
            List of OpenFAST output filepaths or dicts of OpenFAST outputs.
        fatigue_channels : dict (optional)
            Dictionary with format:
            'channel': 'fatigue slope'
        extreme_stat : str (optional)
            Whether the extreme event calculation should work on [max, min, abs].
            Default, 'max'
        extreme_channels : list (optional)
            Limit calculation of extremes to the channel names in the list.
            Unspecified means all channels are processed and reported.
        magnitude_channels : dict (optional)
            Additional channels as vector magnitude of other channels.
            Format: 'new-chan': ['chan1', 'chan2', 'chan3']
        trim_data : tuple or list (optional)
            Trim processed outputs to desired times.
            Format: (min, max)
        """

        self.data        = None
        self.channels    = None
        self.description = kwargs.get("description", "")
        self.units       = kwargs.get("units", None)
        self.filepath    = ""

        for k in ["dlc", "DLC", "name", "Name", "NAME", "filename","filepath","file"]:
            self.filepath = kwargs.get(k, "")
            if len(self.filepath) > 0:
                break

        self.set_data(datain, channelsin)

        self.td = kwargs.get("trim_data", ())
        self.trim_data(*self.td)
        
        self.mc = kwargs.get("magnitude_channels", {})
        self.append_magnitude_channels()

        self.ec = kwargs.get("extreme_channels", [])
        statin  = kwargs.get("extreme_stat", "max")
        self.set_extreme_stat(statin)
        
        self.fc = kwargs.get("fatigue_channels", {})
        
    def __getitem__(self, chan):
        return self.data[:, self.chan_idx(chan)]

    def __str__(self):
        return self.description

    def copy(self):
        return AeroelasticOutput(self.data, self.channels, description=self.description,
                                 units=self.units, name=self.filepath)

    def save(self, fname):
        self.to_df().to_pickle(fname)

    def load(self, fname):
        self.set_data( pd.read_pickle(fname) )

    def set_extreme_stat(self, instr):
        if instr is None or instr == '':
            return
        
        if (not isinstance(instr, str) or
            instr.lower() not in ['max','min','abs','absmax','maximum','minimum','maxima','minima','absolute','abs max']):
            raise ValueError('Expecting stat to be a string of [max, min, abs]')
        
        self.extreme_stat = instr.lower()
        
    def chan_idx(self, chan):
        try:
            idx = self.channels.index(chan)

        except ValueError:
            raise IndexError(f"Channel '{chan}' not found.")
        return idx
    
    def set_data(self, datain, channelsin=None):
        if datain is None:
            return
        
        elif isinstance(datain, dict):
            self.channels = [c.strip() for c in list(datain.keys())]
            self.data     = np.array(list(datain.values())).T
            
        elif isinstance(datain, list):
            self.channels = [c.strip() for c in channelsin]
            self.data     = np.array(datain).T
            
        elif isinstance(datain, pd.DataFrame):
            self.channels = [c.strip() for c in list(datain.columns)]
            self.data     = datain.to_numpy()

        elif isinstance(datain, np.ndarray):
            self.channels = [c.strip() for c in channelsin]
            self.data     = datain

        else:
            pass
            #print("Unknown data type.  Expected dict or list or DataFrame or Numpy Array")
            #print(f"Instead found, {type(datain)}")

    def drop_channel(self, pattern):
        """
        Drop channel based on a string pattern
        """
        idx = np.where([fnmatch.fnmatch(m, pattern) for m in self.channels])[0]
        self.channels = np.delete( np.array(self.channels), idx).tolist()
        self.data = np.delete(self.data, idx, axis=1)
        
    def delete_channel(self, pattern):
        self.drop_channel(pattern)
        
    def remove_channel(self, pattern):
        self.drop_channel(pattern)
        
    def _add_channel(self, datain, namein):
        """
        Add new channels of data.  ASSUMES NUMPY ARRAY
        """
        if namein in self.channels:
            pass
            #print(f"Channel '{namein}' already exists.")
        else:
            self.data = np.c_[self.data, datain]
            self.channels.append( namein )
        
    def add_channel(self, datain, namein=None):
        """
        Add new channels of data.

        Parameters
        ----------
        magnitude_channels : dict
            Format: 'new-chan' ['chan1', 'chan2', 'chan3'],
        """
        if isinstance(datain, dict):
            for c in datain.keys():
                self._add_channel(np.array(datain[c]), c.strip())
            
        elif isinstance(datain, list) and isinstance(datain[0], list) and isinstance(namein, (list, tuple, np.ndarray)):
            for k in range(len(namein)):
                self._add_channel(np.array(datain[k]), namein[k].strip())
            
        elif isinstance(datain, list) and isinstance(datain[0], (int, float)) and isinstance(namein, (list, tuple, np.ndarray)):
            self._add_channel(np.array(datain), namein[0].strip())
            
        elif isinstance(datain, list) and isinstance(namein, str):
            self._add_channel(np.array(datain), namein.strip())
            
        elif isinstance(datain, pd.DataFrame):
            for c in list(datain.columns):
                self._add_channel(datain[c].to_numpy(), c.strip())

        elif isinstance(datain, np.ndarray) and datain.ndim==2 and isinstance(namein, (list, tuple, np.ndarray)):
            for k in range(len(namein)):
                self._add_channel(datain[:,k], namein[k].strip())
                
        elif isinstance(datain, np.ndarray) and datain.ndim==1 and isinstance(namein, (list, tuple, np.ndarray)):
            self._add_channel(datain, namein[0].strip())
                
        elif isinstance(datain, np.ndarray) and isinstance(namein, str):
            self._add_channel(datain, namein.strip())

        elif isinstance(datain, str) and isinstance(namein, str):
            new_data = ne.evaluate(datain, local_dict=self.to_dict())
            self._add_channel(new_data, namein.strip())

        else:
            print("Unknown data type.  Expected dict or list or DataFrame or Numpy Array")
            print(f"Instead found, {type(datain)}")


    def calculate_channel(self, instr, namein):
        self.add_channel(instr, namein)
        
    def to_df(self):
        """Returns `self.data` as a DataFrame."""

        if self.channels is None:
            return pd.DataFrame(self.data)
        else:
            return pd.DataFrame(self.data, columns=self.channels)

    @dataproperty
    def df(self):
        return self.to_df()
        
    def to_dict(self):
        """Returns `self.data` as a dictionary."""

        return {self.channels[k]:self.data[:,k].tolist() for k in range(len(self.channels))}

    @dataproperty
    def time(self):
        if "Time" in self.channels:
            return self["Time"]
        else:
            return self.data[:,0]

    @property
    def filename(self):
        return os.path.split(self.filepath)[-1]
                
    @property
    def headers(self):
        if getattr(self, "units", None) is None:
            return None

        else:
            return [
                self.channels[k]
                if self.units[k] in ["", "-"]
                else f"{self.channels[k]} [{self.units[k]}]"
                for k in range(self.num_channels)
            ]
            
    def trim_data(self, tmin=0, tmax=np.inf):
        """
        Trims `self.data` to the data between `tmin` and `tmax`.

        Parameters
        ----------
        tmin : int | float
            Start time.
        tmax : int | float
            Ending time.
        """
        if tmin is None:
            tmin = 0.0
        if tmax is None:
            tmax = np.inf
        if tmin==0.0 and tmax==np.inf:
            return
        
        idx = np.where((self.time >= tmin) & (self.time <= tmax))[0]
        if len(idx) == 0:
            raise ValueError(
                f"No time steps after trimming. Provided: ({tmin}, {tmax}). Available: ({self.time.min()}, {self.time.max()})"
            )

        self.data = self.data[idx,:]

    def add_gradient_channel(self, chanstr, new_name):
        newdata = np.gradient(self[chanstr], self.time)
        self._add_channel(newdata, new_name)
        
    def append_magnitude_channels(self, magnitude_channels=None):
        """
        Append the vector magnitude of `channels` to the dataset.

        Parameters
        ----------
        magnitude_channels : dict
            Format: 'new-chan' ['chan1', 'chan2', 'chan3'],
        """

        if magnitude_channels is not None:
            if not isinstance(magnitude_channels, dict):
                raise ValueError("Expecting magnitude channels as a dictionary, 'new-chan': ['chan1', 'chan2', 'chan3']")
            
            self.mc.update( magnitude_channels)
        
        if self.mc is None:
            return
        
        for new_chan, chans in self.mc.items():

            try:
                arrays = np.array([self[a] for a in chans]).T
                normed = np.linalg.norm(arrays, axis=1).reshape(arrays.shape[0], 1)
            except:
                normed = np.nan * np.ones((self.data.shape[0],1))

            self._add_channel(normed, new_chan)
            
    def add_magnitude_channels(self, magnitude_channels=None):
        """
        Since add/append get confused quite a bit
        """
        self.append_magnitude_channels(magnitude_channels=magnitude_channels)

    def add_load_rose(self, load_rose=None, nsec=6):
        """
        Append a load rose of `nsec` sectors of x-y `channels` to the dataset.
        The x-channel will be multiplied by cos(theta) and y-channel by sin(theta)

        Parameters
        ----------
        load_rose : dict
            Format: 'new-chan' ['chan_x', 'chan_y'],
        """

        if load_rose is None:
            return
        else:
            if not isinstance(load_rose, dict):
                raise ValueError("Expecting load rose channels as a dictionary, 'new-chan': ['chan_x', 'chan_y']")

        th_pts = np.linspace(0, 180, nsec+1) # Don't need to repeat the 360 mark
        thd = 0.5*(th_pts[:-1] + th_pts[1:]) # Mid point of each sector
        
        for rootstr, chans in load_rose.items():
            for td in thd:
                tr = np.deg2rad(td)
                new_chan = rootstr + str(int(td))
                new_data = self[chans[0]]*np.cos(tr) + self[chans[1]]*np.sin(tr)
                self._add_channel(new_data, new_chan)
                print(f"Added channel, {new_chan}")
        
    @dataproperty
    def num_timesteps(self):
        return self.data.shape[0]
    
    @dataproperty
    def dt(self):
        return self.time[1] - self.time[0]

    @dataproperty
    def elapsed_time(self):
        return self.time.max() - self.time.min()

    @dataproperty
    def num_channels(self):
        return self.data.shape[1]

    @dataproperty
    def idxmins(self):
        return self.data.argmin(axis=0)

    @dataproperty
    def idxmaxs(self):
        return self.data.argmax(axis=0)

    @dataproperty
    def idxabs(self):
        return np.abs(self.data).argmax(axis=0)

    @dataproperty
    def minima(self):
        return self.data.min(axis=0)

    @dataproperty
    def maxima(self):
        return self.data.max(axis=0)

    @dataproperty
    def ranges(self):
        return self.maxima - self.minima

    @dataproperty
    def variable(self):
        return np.where(self.ranges != 0.0)[0]

    @dataproperty
    def constant(self):
        return np.where(self.ranges == 0.0)[0]

    @dataproperty
    def sums(self):
        return self.data.sum(axis=0)

    @dataproperty
    def sums_squared(self):
        return np.sum(self.data ** 2, axis=0)

    @dataproperty
    def sums_cubed(self):
        return np.sum(self.data ** 3, axis=0)

    @dataproperty
    def sums_fourth(self):
        return np.sum(self.data ** 4, axis=0)

    @dataproperty
    def second_moments(self):
        return stats.moment(self.data, moment=2, axis=0)

    @dataproperty
    def third_moments(self):
        return stats.moment(self.data, moment=3, axis=0)

    @dataproperty
    def fourth_moments(self):
        return stats.moment(self.data, moment=4, axis=0)

    @dataproperty
    def means(self):
        return self.data.mean(axis=0)

    @dataproperty
    def medians(self):
        return np.median(self.data, axis=0)

    @dataproperty
    def absmaxima(self):
        return np.abs(self.data).max(axis=0)

    @dataproperty
    def stddevs(self):
        return self.data.std(axis=0)

    @dataproperty
    def skews(self):
        return self.third_moments / np.sqrt(self.second_moments) ** 3

    @dataproperty
    def kurtosis(self):
        return self.fourth_moments / self.second_moments ** 2

    @dataproperty
    def integrated(self):
        return np.trapezoid(self.data, self.time, axis=0)

    def compute_energy(self, pwrchan):
        return np.trapezoid(self[pwrchan], self.time)

    def total_travel(self, chanstr):
        dchan = np.gradient(self[chanstr], self.time)
        return np.trapezoid(np.abs(dchan), self.time)

    def histogram(self, chanstr, bins=15):
        return np.histogram(self[chanstr], bins=bins, density=False)
    
    def density(self, chanstr, bins=15):
        return np.histogram(self[chanstr], bins=bins, density=True)
    
    def psd(self, nfft=None):
        """
        Compute power spectra density for each channel.

        Returns
        ----------
        freq : Numpy array 1-D
            Valid frequencies for spectra
        Pxx_den : Numpy array 2-D
            Power spectral density with each column corresponding to a channel
        """
        fs = 1. / np.diff(self.time)[0]
        freq, Pxx_den = signal.welch(self.data, fs, nfft=nfft, axis=0)
        fobj = self.copy()
        fobj.data = Pxx_den
        fobj.data[:,0] = freq
        fobj.channels[0] = 'Freq'
        return fobj

    def time_averaging(self, time_window):
        """
        Applies averaging/smoothing window on time-domain channels via convolution.
        Window input is seconds
        
        Parameters
        ----------
        time_windows : float
            Window for averaging/smoothing in seconds

        Returns
        ----------
        obj : AeroelasticOutput object
            New AeroelasticOutput instance with the new data
        """
        dt = np.diff(self.time)[0]
        npts = int(time_window / dt)
        window = np.ones(npts) / npts # Basic rectangular filter

        new_timesteps = max(npts,self.num_timesteps) - min(npts,self.num_timesteps) + 1
        data_avg = np.zeros((new_timesteps, self.num_channels))
        for k in range(len(self.channels)):
            data_avg[:,k] = np.convolve(self.data[:,k], window, 'valid')

        # Install the new data
        self.data = data_avg
        
    def averaging(self, window):
        self.time_averaging(window)

    
    def time_binning(self, time_window):
        """
        Bin the data into groups specified by time_window (in seconds)
        Average (or, someday, other stats of those windows)
        
        Parameters
        ----------
        time_windows : float
            Window for averaging/smoothing in seconds

        Returns
        ----------
        obj : AeroelasticOutput object
            New AeroelasticOutput instance with the new data
        """

        n_bins = int(np.ceil((self.time.max() - self.time.min()) / time_window))

        data_binned = np.zeros((n_bins,self.data.shape[1]))

        for i_bin in range(n_bins):
            t_start = i_bin * time_window  + self.time.min()
            t_end = (i_bin+1) * time_window    + self.time.min()
            time_index = np.logical_and(self.time >= t_start, self.time < t_end)
            if not np.any(time_index):
                raise Exception("Error binning time between {t_start} and {t_end}")
            data_filtered = self.data[time_index,:]
            data_binned[i_bin,:] = np.mean(data_filtered,axis=0)

        # Install the new data
        self.data = data_binned
        
    def binning(self, window):
        self.time_binning(window)

    def summary_stats(self):
        """
        Appends summary statistics to `self.summary_statistics` for each file.
        """

        mins = self.minima
        maxs = self.maxima
        stds = self.stddevs
        means = self.means
        meds  = self.medians
        abss  = self.absmaxima
        intgs = self.integrated
        
        fstats = {}
        for k in range(len(self.channels)):
            channel = self.channels[k]
            
            fstats[channel] = {
                "min": mins[k],
                "max": maxs[k],
                "std": stds[k],
                "mean": means[k],
                "median": meds[k],
                "abs": abss[k],
                "integrated": intgs[k],
            }

        return fstats

    
    def extremes(self, chanlist=None, stat=None):
        """"""
        if chanlist is not None:
            if not isinstance(chanlist, (list, set, tuple)):
                raise ValueError("Expecting extreme channels as a list, tuple, or set")
            
            self.ec = list(set(self.ec + chanlist))
        
        if self.ec is None or len(self.ec) == 0:
            channels = self.channels
        else:
            channels = self.ec

        sorter = np.argsort(self.channels)
        exists = [c for c in channels if c in self.channels]
        idx = sorter[np.searchsorted(self.channels, exists, sorter=sorter)]

        # Get statistic
        self.set_extreme_stat(stat)
        if self.extreme_stat.find('abs') >= 0:
            mystat = self.idxabs
        elif self.extreme_stat.find('min') >= 0:
            mystat = self.idxmins
        elif self.extreme_stat.find('max') >= 0:
            mystat = self.idxmaxs
        
        extremes = {}
        for chan, i in zip(exists, idx):
            idx_stat = mystat[i]
            extremes[chan] = {
                "Time": self.time[idx_stat],
                **dict(zip(exists, self.data[idx_stat, idx])),
            }

        return extremes


        
    def get_DELs(self, **kwargs):
        """
        Appends computed damage equivalent loads for fatigue channels in
        `self.fc`.

        Parameters
        ----------
        output : AerolelasticOutput
        rainflow_bins : int (optional)
            Number of bins used in rainflow analysis.
            Default: 100
        goodman_correction: boolean (optional)
            Whether to apply Goodman mean correction to loads and stress
            Default: False
        """
        for k in kwargs:
            if k not in ["rainflow_bins", "bins",
                         "goodman_correction", "goodman"]:
                print(f"Unknown keyword argument, {k}")

        DELs = {}
        D = {}

        for chan, fatparams in self.fc.items():
            goodman = kwargs.get("goodman_correction", fatparams.goodman)
            goodman = kwargs.get("goodman", goodman)
            bins = kwargs.get("rainflow_bins", fatparams.bins)
            bins = kwargs.get("bins", bins)
                
            try:
                if np.isnan(np.sum(self[chan])):  # channel has a nan
                    raise IndexError

                DELs[chan] = fatparams.compute_del(self[chan], self.elapsed_time,
                                                   goodman_correction=goodman,
                                                   rainflow_bins=bins)
                
                if np.abs(fatparams.load2stress) > 0.0:
                    D[chan] = fatparams.compute_damage(self[chan],
                                                       goodman_correction=goodman,
                                                       rainflow_bins=bins)
                else:
                    D[chan] = 0.0

            except IndexError:
                print(f"Channel '{chan}' not included in DEL calculation.")
                DELs[chan] = np.nan
                D[chan] = np.nan

        return DELs, D

    
    def process(self, **kwargs):
        """
        Process AeroelasticOutput output for summary stats, extreme event table, DELs, and Damage
        """

        # Data manipulation list all other outputs
        self.trim_data(*self.td)
        self.append_magnitude_channels(self.mc)
        
        self.stats = self.summary_stats()

        self.ext_table = self.extremes()

        self.dels, self.damage = self.get_DELs(**kwargs)




import os
import unittest
import importlib
from time import time
from pathlib import Path

thisdir = os.path.dirname(os.path.realpath(__file__))
# Find examples directory- outside the module path

root_dir = os.path.dirname(os.path.dirname(thisdir))
examples_dir = os.path.join(root_dir, "examples")
all_examples = Path(examples_dir).rglob("*.py") if os.path.exists(examples_dir) else []

def execute_script(fscript):
    # Go to location due to relative path use for airfoil files
    print("\n\n")
    print("NOW RUNNING:", fscript)
    print()
    fullpath = os.path.join(examples_dir, fscript)
    basepath = os.path.join(examples_dir, str(fscript).split(os.sep)[0])
    os.chdir(basepath)

    # Get script/module name
    froot = str(fscript).split(os.sep)[-1]

    # Use dynamic import capabilities
    # https://www.blog.pythonlibrary.org/2016/05/27/python-201-an-intro-to-importlib/
    print(froot, os.path.realpath(fullpath))
    spec = importlib.util.spec_from_file_location(froot, os.path.realpath(fullpath))
    mod = importlib.util.module_from_spec(spec)
    s = time()
    spec.loader.exec_module(mod)
    print(time() - s, "seconds to run")


class TestExamples(unittest.TestCase):
    def test_all_scripts(self):
        for ks, s in enumerate(all_examples):
            with self.subTest(f"Running: {s}", i=ks):
                try:
                    print(s)
                    execute_script(s)
                    self.assertTrue(True)
                except Exception:
                    self.assertEqual(s, "Success")

    
if __name__ == "__main__":
    unittest.main()
    



import os
import unittest
import numpy as np
import numpy.testing as npt
import pandas as pd
#import pandas.testing as pdt

from pCrunch import AeroelasticOutput

data = {
    "Time": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
    "WindVxi": [7, 7, 7, 7, 7, 8, 8, 8, 8, 8],
    "WindVyi": [0] * 10,
    "WindVzi": [0] * 10,
}
mc = {"Wind": ["WindVxi", "WindVyi", "WindVzi"]}

class Test_AeroelasticOutput(unittest.TestCase):

    def testConstructor(self):
        # Empty
        myobj = AeroelasticOutput()
        self.assertEqual(myobj.data, None)
        self.assertEqual(myobj.channels, None)
        self.assertEqual(myobj.units, None)
        self.assertEqual(myobj.description, "")
        self.assertEqual(myobj.filepath, "")
        self.assertEqual(myobj.extreme_stat, "max")
        self.assertEqual(myobj.td, ())
        self.assertEqual(myobj.mc, {})
        self.assertEqual(myobj.ec, [])
        self.assertEqual(myobj.fc, {})

        # Test keywords
        myobj = AeroelasticOutput(description="test", units=["m"]*5, name="label", extreme_stat='abs')
        self.assertEqual(myobj.data, None)
        self.assertEqual(myobj.channels, None)
        self.assertEqual(myobj.units, ["m"]*5)
        self.assertEqual(myobj.description, "test")
        self.assertEqual(myobj.filepath, "label")
        self.assertEqual(myobj.extreme_stat, "abs")
        self.assertEqual(myobj.td, ())
        self.assertEqual(myobj.mc, {})
        self.assertEqual(myobj.ec, [])
        self.assertEqual(myobj.fc, {})

        # As dict
        myobj = AeroelasticOutput(data, magnitude_channels=mc)
        self.assertEqual(myobj.data.shape, (10,5))
        npt.assert_equal(myobj.data[:,0], np.array(data["Time"]))
        npt.assert_equal(myobj.data[:,1], np.array(data["WindVxi"]))
        npt.assert_equal(myobj.data[:,2], np.zeros(10))
        npt.assert_equal(myobj.data[:,3], np.zeros(10))
        npt.assert_equal(myobj.data[:,4], np.array(data["WindVxi"]))
        self.assertEqual(myobj.channels, list(data.keys())+["Wind"])
        self.assertEqual(myobj.units, None)
        self.assertEqual(myobj.description, "")
        self.assertEqual(myobj.filepath, "")
        self.assertEqual(myobj.extreme_stat, "max")
        self.assertEqual(myobj.td, ())
        self.assertEqual(myobj.mc, mc)
        self.assertEqual(myobj.ec, [])
        self.assertEqual(myobj.fc, {})

        # As DataFrame
        myobj = AeroelasticOutput(pd.DataFrame(data), magnitude_channels=mc)
        self.assertEqual(myobj.data.shape, (10,5))
        npt.assert_equal(myobj.data[:,0], np.array(data["Time"]))
        npt.assert_equal(myobj.data[:,1], np.array(data["WindVxi"]))
        npt.assert_equal(myobj.data[:,2], np.zeros(10))
        npt.assert_equal(myobj.data[:,3], np.zeros(10))
        npt.assert_equal(myobj.data[:,4], np.array(data["WindVxi"]))
        self.assertEqual(myobj.channels, list(data.keys())+["Wind"])
        self.assertEqual(myobj.units, None)
        self.assertEqual(myobj.description, "")
        self.assertEqual(myobj.filepath, "")
        self.assertEqual(myobj.extreme_stat, "max")
        self.assertEqual(myobj.td, ())
        self.assertEqual(myobj.mc, mc)
        self.assertEqual(myobj.ec, [])
        self.assertEqual(myobj.fc, {})

        # As Numpy
        myobj = AeroelasticOutput(myobj.data[:,:-1], list(data.keys()), magnitude_channels=mc)
        self.assertEqual(myobj.data.shape, (10,5))
        npt.assert_equal(myobj.data[:,0], np.array(data["Time"]))
        npt.assert_equal(myobj.data[:,1], np.array(data["WindVxi"]))
        npt.assert_equal(myobj.data[:,2], np.zeros(10))
        npt.assert_equal(myobj.data[:,3], np.zeros(10))
        npt.assert_equal(myobj.data[:,4], np.array(data["WindVxi"]))
        self.assertEqual(myobj.channels, list(data.keys())+["Wind"])
        self.assertEqual(myobj.units, None)
        self.assertEqual(myobj.description, "")
        self.assertEqual(myobj.filepath, "")
        self.assertEqual(myobj.extreme_stat, "max")
        self.assertEqual(myobj.td, ())
        self.assertEqual(myobj.mc, mc)
        self.assertEqual(myobj.ec, [])
        self.assertEqual(myobj.fc, {})

        # As lists
        myobj = AeroelasticOutput([m for m in data.values()], list(data.keys()), magnitude_channels=mc)
        self.assertEqual(myobj.data.shape, (10,5))
        npt.assert_equal(myobj.data[:,0], np.array(data["Time"]))
        npt.assert_equal(myobj.data[:,1], np.array(data["WindVxi"]))
        npt.assert_equal(myobj.data[:,2], np.zeros(10))
        npt.assert_equal(myobj.data[:,3], np.zeros(10))
        npt.assert_equal(myobj.data[:,4], np.array(data["WindVxi"]))
        self.assertEqual(myobj.channels, list(data.keys())+["Wind"])
        self.assertEqual(myobj.units, None)
        self.assertEqual(myobj.description, "")
        self.assertEqual(myobj.filepath, "")
        self.assertEqual(myobj.extreme_stat, "max")
        self.assertEqual(myobj.td, ())
        self.assertEqual(myobj.mc, mc)
        self.assertEqual(myobj.ec, [])
        self.assertEqual(myobj.fc, {})

        # With trimming
        myobj = AeroelasticOutput(data, magnitude_channels=mc, trim_data=[3,6])
        self.assertEqual(myobj.data.shape, (4,5))
        npt.assert_equal(myobj.data[:,0], np.array(data["Time"][2:6]))
        npt.assert_equal(myobj.data[:,1], np.array(data["WindVxi"][2:6]))
        npt.assert_equal(myobj.data[:,2], np.zeros(4))
        npt.assert_equal(myobj.data[:,3], np.zeros(4))
        npt.assert_equal(myobj.data[:,4], np.array(data["WindVxi"][2:6]))
        self.assertEqual(myobj.channels, list(data.keys())+["Wind"])
        self.assertEqual(myobj.units, None)
        self.assertEqual(myobj.description, "")
        self.assertEqual(myobj.filepath, "")
        self.assertEqual(myobj.extreme_stat, "max")
        self.assertEqual(myobj.td, [3, 6])
        self.assertEqual(myobj.mc, mc)
        self.assertEqual(myobj.ec, [])
        self.assertEqual(myobj.fc, {})
        
    def testGetters(self):
        myobj = AeroelasticOutput(data, magnitude_channels=mc, dlc="/testdir/testfile")
        npt.assert_equal(myobj.time, data["Time"])
        npt.assert_equal(myobj.filepath, "/testdir/testfile")
        npt.assert_equal(myobj.filename, "testfile")
        for k in data.keys():
            npt.assert_equal(myobj[k], data[k])
        npt.assert_equal(myobj["Wind"], data["WindVxi"])
        self.assertEqual(myobj.chan_idx("WindVxi"), 1)

        mydict = myobj.to_dict()
        for k in data.keys():
            npt.assert_equal(mydict[k], data[k])

    def testAddDeleteChannel(self):
        # Adding same ones
        myobj = AeroelasticOutput(data, magnitude_channels=mc)
        self.assertEqual(myobj.data.shape, (10,5))
        myobj.add_channel(data) # SHouldn't add anything
        self.assertEqual(myobj.data.shape, (10,5))
        npt.assert_equal(myobj.data[:,0], np.array(data["Time"]))
        npt.assert_equal(myobj.data[:,1], np.array(data["WindVxi"]))
        npt.assert_equal(myobj.data[:,2], np.zeros(10))
        npt.assert_equal(myobj.data[:,3], np.zeros(10))
        self.assertEqual(myobj.channels, list(data.keys())+["Wind"])

        # Add in different types
        myobj.add_channel({"Test1":data["WindVxi"]})
        myobj.add_channel(pd.DataFrame({"Test2":data["WindVxi"]}))
        myobj.add_channel(data["WindVxi"], "Test3")
        myobj.add_channel(data["WindVxi"], ["Test4"])
        myobj.add_channel(data["WindVxi"], ("Test5",))
        myobj.add_channel(np.array(data["WindVxi"]), "Test6")
        myobj.add_channel(np.array(data["WindVxi"]), ["Test7"])
        self.assertEqual(myobj.data.shape, (10,12))
        for k in range(5,12):
            npt.assert_equal(myobj.data[:,k], np.array(data["WindVxi"]))

        # Add gradient
        myobj.add_gradient_channel('WindVxi', 'Test8')
        npt.assert_almost_equal(myobj['Test8'], np.r_[np.zeros(4), 0.5, 0.5, np.zeros(4)])
            
        # Now delete them
        myobj.drop_channel('Test*')
        self.assertEqual(myobj.data.shape, (10,5))

    def testCalcChannel(self):
        myobj = AeroelasticOutput(data, magnitude_channels=mc)
        myobj.calculate_channel("2*WindVxi", "Test1")
        myobj.calculate_channel("WindVxi + WindVyi", "Test2")
        myobj.calculate_channel("WindVxi - WindVzi", "Test3")
        myobj.calculate_channel("WindVxi**2", "Test4")
        self.assertEqual(myobj.data.shape, (10,9))
        npt.assert_equal(myobj.data[:,5], 2*np.array(data["WindVxi"]))
        npt.assert_equal(myobj.data[:,6], np.array(data["WindVxi"]))
        npt.assert_equal(myobj.data[:,7], np.array(data["WindVxi"]))
        npt.assert_equal(myobj.data[:,8], np.array(data["WindVxi"])**2)

    def testTrim(self):
        myobj = AeroelasticOutput(data, magnitude_channels=mc)
        myobj.trim_data(3, 6)
        self.assertEqual(myobj.data.shape, (4,5))
        npt.assert_equal(myobj.time, np.arange(3,7))

    def testLoadRose(self):
        myobj = AeroelasticOutput(data, magnitude_channels=mc)
        myobj.add_load_rose({"WindLR":["WindVxi", "WindVyi"]})
        self.assertEqual(myobj.data.shape, (10,11))
        npt.assert_equal(myobj.data[:,5], np.array(data["WindVxi"])*np.cos(np.deg2rad(15)))
        npt.assert_equal(myobj.data[:,6], np.array(data["WindVxi"])*np.cos(np.deg2rad(45)))
        npt.assert_equal(myobj.data[:,7], np.array(data["WindVxi"])*np.cos(np.deg2rad(75)))
        npt.assert_equal(myobj.data[:,8], np.array(data["WindVxi"])*np.cos(np.deg2rad(105)))
        npt.assert_equal(myobj.data[:,9], np.array(data["WindVxi"])*np.cos(np.deg2rad(135)))
        npt.assert_equal(myobj.data[:,10], np.array(data["WindVxi"])*np.cos(np.deg2rad(165)))

    def testProperties(self):
        myobj = AeroelasticOutput(data, magnitude_channels=mc)
        npt.assert_equal(myobj.num_timesteps, 10)
        npt.assert_equal(myobj.dt, 1)
        npt.assert_equal(myobj.elapsed_time, 9)
        npt.assert_equal(myobj.num_channels, 5)
        npt.assert_equal(myobj.idxmins, np.zeros(5))
        npt.assert_equal(myobj.idxmaxs, [9, 5, 0, 0, 5])
        npt.assert_equal(myobj.minima, [1, 7, 0, 0, 7])
        npt.assert_equal(myobj.maxima, [10, 8, 0, 0, 8])
        npt.assert_equal(myobj.absmaxima, [10, 8, 0, 0, 8])
        npt.assert_equal(myobj.ranges, [9, 1, 0, 0, 1])
        npt.assert_equal(myobj.variable, [0, 1, 4])
        npt.assert_equal(myobj.constant, [2, 3])
        npt.assert_equal(myobj.sums.shape, (myobj.num_channels,))
        npt.assert_equal(myobj.sums_squared.shape, (myobj.num_channels,))
        npt.assert_equal(myobj.sums_cubed.shape, (myobj.num_channels,))
        npt.assert_equal(myobj.sums_fourth.shape, (myobj.num_channels,))
        npt.assert_equal(myobj.second_moments.shape, (myobj.num_channels,))
        npt.assert_equal(myobj.third_moments.shape, (myobj.num_channels,))
        npt.assert_equal(myobj.fourth_moments.shape, (myobj.num_channels,))
        npt.assert_equal(myobj.means.shape, (myobj.num_channels,))
        npt.assert_equal(myobj.medians.shape, (myobj.num_channels,))
        npt.assert_equal(myobj.stddevs.shape, (myobj.num_channels,))
        npt.assert_equal(myobj.skews.shape, (myobj.num_channels,))
        npt.assert_equal(myobj.kurtosis.shape, (myobj.num_channels,))
        npt.assert_equal(myobj.integrated.shape, (myobj.num_channels,))
        npt.assert_equal(myobj.compute_energy('WindVxi'), myobj.integrated[1])
        npt.assert_equal(myobj.total_travel('WindVxi'), 1.0)

    def test_windowing(self):
        myobj = AeroelasticOutput(data, magnitude_channels=mc)
        myobj.time_averaging(2.0)
        npt.assert_equal(myobj.data[:,0], np.array(data["Time"][:-1]) + 0.5)
        npt.assert_equal(myobj.data[:,2], 0.0)
        npt.assert_equal(myobj.data[:,3], 0.0)

        myobj = AeroelasticOutput(data, magnitude_channels=mc)
        myobj.time_binning(2.0)
        npt.assert_equal(myobj.data[:,0], np.array(data["Time"][::2]) + 0.5)
        npt.assert_equal(myobj.data[:,2], 0.0)
        npt.assert_equal(myobj.data[:,3], 0.0)

    def test_psd(self):
        myobj = AeroelasticOutput(data, magnitude_channels=mc)
        freq_obj = myobj.psd()
        self.assertEqual(myobj.num_channels, freq_obj.num_channels)
        
    def test_stats_extremes(self):
        myobj = AeroelasticOutput(data, magnitude_channels=mc)

        stats = myobj.summary_stats()
        self.assertEqual(stats["Time"]['min'], 1.)
        self.assertEqual(stats["Time"]['max'], 10.)
        self.assertEqual(stats["Time"]['abs'], 10.)
        self.assertEqual(stats["Time"]['mean'], 5.5)
        self.assertEqual(stats["Time"]['median'], 5.5)
        self.assertEqual(stats["Time"]['integrated'], 49.5)

        ext = myobj.extremes()
        self.assertEqual(ext["Time"], {"Time":10., "WindVxi":8., "WindVyi":0., "WindVzi":0, "Wind":8.})

        ext = myobj.extremes(stat='min')
        self.assertEqual(ext["Time"], {"Time":1., "WindVxi":7., "WindVyi":0., "WindVzi":0, "Wind":7.})
        self.assertEqual(myobj.extreme_stat, "min")

        ext = myobj.extremes(stat='absmax')
        self.assertEqual(ext["Time"], {"Time":10., "WindVxi":8., "WindVyi":0., "WindVzi":0, "Wind":8.})
        self.assertEqual(myobj.extreme_stat, "absmax")

        
    def test_process(self):
        myobj = AeroelasticOutput(data, magnitude_channels=mc)
        
        stats = myobj.summary_stats()
        ext_tab = myobj.extremes()
        self.assertFalse(hasattr(myobj, 'stats'))

        myobj.process()
        self.assertTrue(hasattr(myobj, 'stats'))
        self.assertEqual(myobj.stats, stats)
        self.assertEqual(myobj.ext_table["Time"], ext_tab["Time"])
        
if __name__ == "__main__":
    unittest.main()



import numpy as np
from .aeroelastic_output import AeroelasticOutput

def read(filename, **kwargs):
    """
    Load a single OpenFAST file.

    Parameters
    ----------
    filename : Path object or string
        OpenFAST file to load.

    Returns
    -------
    output
        OpenFASTOutput instances
    """

    batch_flag = isinstance(filename, (list, tuple, set))
    filelist = filename if batch_flag else [filename]
    
    fastout = []
    for k in filelist:
        try:
            output = OpenFASTAscii(k, **kwargs)
            output.read()
            
        except UnicodeDecodeError:
            output = OpenFASTBinary(k, **kwargs)
            output.read()
            
        fastout.append(output)

    if batch_flag:
        return fastout
    else:
        return fastout[0]
    
def read_parallel(filenames, n_cores=1, **kwargs):
    """
    Read OpenFAST files in parallel.

    Parameters
    ----------
    filenames : list of str
        List of OpenFAST file paths.
    n_cores : int, optional
        Number of cores to use for parallel reading. Default is 1.

    Returns
    -------
    list of OpenFASTOutput
        List of OpenFASTOutput instances.
    """
    
    import multiprocessing
    with multiprocessing.Pool(n_cores) as pool:
        outputs = pool.map(read, filenames, **kwargs)
    
    return outputs

    
def load_FAST_out(filenames, tmin=0, tmax=float('inf'), **kwargs):
    """ Backwards compatibility with old name """
    fastout = read(filenames, tmin=tmin, tmax=tmax, **kwargs)
    for k in fastout:
        k.trim_data(tmin, tmax)
    return fastout



class OpenFASTBinary(AeroelasticOutput):
    """OpenFAST binary output class."""

    def __init__(self, filepath, **kwargs):
        """
        Creates an instance of `OpenFASTBinary`.

        Parameters
        ----------
        filepath : path-like
        """

        super().__init__()
        self.filepath = filepath
        self._chan_chars = kwargs.get("chan_char_length", 10)
        self._unit_chars = kwargs.get("unit_char_length", 10)
        mc = kwargs.get("magnitude_channels", None)
        self.read(magnitude_channels=mc)
        self.fmt = 0

    def read(self, magnitude_channels=None):
        """Reads the binary file."""

        with open(self.filepath, "rb") as f:
            self.fmt = np.fromfile(f, np.int16, 1)[0]

            if self.fmt == 4:
                self._chan_chars = np.fromfile(f, np.int16, 1)[0]
                self._unit_chars = self._chan_chars

            num_channels = np.fromfile(f, np.int32, 1)[0]
            num_timesteps = np.fromfile(f, np.int32, 1)[0]
            num_points = num_channels * num_timesteps
            time_info = np.fromfile(f, np.float64, 2)

            if self.fmt == 3:
                slopes = np.ones(num_channels)
                offset = np.zeros(num_channels)

            else:
                slopes = np.fromfile(f, np.float32, num_channels)
                offset = np.fromfile(f, np.float32, num_channels)

            length = np.fromfile(f, np.int32, 1)[0]
            chars = np.fromfile(f, np.uint8, length)
            self.description = "".join(map(chr, chars)).strip()

            # Build headers
            channels = np.fromfile(
                f, np.uint8, self._chan_chars * (num_channels + 1)
            ).reshape((num_channels + 1), self._chan_chars)
            channels_list = list("".join(map(chr, c)) for c in channels)
            self.channels = [c.strip() for c in channels_list]

            units = np.fromfile(
                f, np.uint8, self._unit_chars * (num_channels + 1)
            ).reshape((num_channels + 1), self._unit_chars)
            self.units = list("".join(map(chr, c)).strip()[1:-1] for c in units)
            
            # Build time
            if self.fmt == 1:
                scale, offset = time_info
                data = np.fromfile(f, np.int32, num_timesteps)
                time = (data - offset) / scale

            else:
                t1, incr = time_info
                time = t1 + incr * np.arange(num_timesteps)

            # Build the data
            if self.fmt == 3:
                raw = np.fromfile(f, np.float64, count=num_points).reshape(num_timesteps, num_channels)
                self.data = np.concatenate([time.reshape(num_timesteps, 1),
                                            raw], 1)

            else:
                raw = np.fromfile(f, np.int16, count=num_points).reshape(num_timesteps, num_channels)
                self.data = np.concatenate([time.reshape(num_timesteps, 1),
                                            (raw - offset) / slopes],
                                           1)

        self.append_magnitude_channels(magnitude_channels=magnitude_channels)


class OpenFASTAscii(AeroelasticOutput):
    """
    OpenFAST ASCII output class.

    Built using data from the Mlife post-processor scripts. May not emcompass
    new OpenFAST output formats.
    """

    def __init__(self, filepath, **kwargs):
        """
        Creates an instance of `OpenFASTAscii`.

        Parameters
        ----------
        filepath : path-like
        """

        super().__init__()
        self.filepath = filepath
        mc = kwargs.get("magnitude_channels", None)
        self.read(magnitude_channels=mc)

    def read(self, magnitude_channels=None):
        """Reads the ASCII file."""

        with open(self.filepath, "rb") as f:
            chandata, unitdata = self.parse_header(f)
            self.build_headers(chandata, unitdata)
            self.data = np.fromfile(f, float, sep="\t").reshape(
                -1, len(self.channels)
            )

        self.append_magnitude_channels(magnitude_channels=magnitude_channels)

    def parse_header(self, f):
        """Reads the header data for file."""

        _start = False
        data = []

        for _line in f:
            try:
                line = _line.replace(b"\xb7", b"-").decode(encoding="ascii").strip()
            except Exception:
                line = _line.replace(b"\xb7", b"-").decode(encoding="utf-16").strip()
            data.append(line)

            if _start:
                break

            if line.startswith("Time"):
                _start = True

        self.description = " ".join([h.replace('"', "") for h in data[:-2]]).strip()

        chandata, unitdata = data[-2:]
        return chandata, unitdata

    def build_headers(self, chandata, unitdata):
        """Unpacks channels and units and builds the combined headers."""

        self.channels = [c.strip() for c in chandata.split("\t")]
        self.units = [u[1:-1] for u in unitdata.split("\t")]

        



from .crunch import Crunch
from .fatigue import FatigueParams
from .openfast_readers import OpenFASTAscii, OpenFASTBinary, load_FAST_out, read, read_parallel
from .aeroelastic_output import AeroelasticOutput



import os
import unittest
import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt

from pCrunch import AeroelasticOutput, FatigueParams, read

DIR = os.path.split(__file__)[0]
DATA = os.path.join(DIR, "data")

FOUT  = ['AOC_WSt.out', 'DLC2.3_1.out', 'DLC2.3_2.out', 'DLC2.3_3.out']

class Test_Fatigue(unittest.TestCase):

    def test_init(self):
        # Test DNV curves
        myfat = FatigueParams(dnv_type='Air', dnv_name='d')
        self.assertEqual(myfat.load2stress, 1.0)
        self.assertEqual(myfat.S_ult, 1.0)
        self.assertEqual(myfat.bins, 100.0)
        self.assertEqual(myfat.goodman, False)
        self.assertEqual(myfat.curve.m1, 3.0)
        self.assertEqual(myfat.curve.m2, 5.0)

        myfat = FatigueParams(load2stress=25.0, dnv_type='sea', dnv_name='c1')
        self.assertEqual(myfat.load2stress, 25.0)
        self.assertEqual(myfat.S_ult, 1.0)
        self.assertEqual(myfat.bins, 100.0)
        self.assertEqual(myfat.goodman, False)
        self.assertEqual(myfat.curve.m, 3.0)

        myfat = FatigueParams(bins=256, goodman=True, ultimate_stress=1e6, dnv_type='cathodic', dnv_name='b2')
        self.assertEqual(myfat.load2stress, 1.0)
        self.assertEqual(myfat.S_ult, 1e6)
        self.assertEqual(myfat.bins, 256.0)
        self.assertEqual(myfat.goodman, True)
        self.assertEqual(myfat.curve.m1, 4.0)
        self.assertEqual(myfat.curve.m2, 5.0)

        # Test Sc-Nc
        myfat = FatigueParams(Sc=1e9, Nc=2e6, slope=3)
        self.assertEqual(myfat.load2stress, 1.0)
        self.assertEqual(myfat.S_ult, 1.0)
        self.assertEqual(myfat.bins, 100.0)
        self.assertEqual(myfat.goodman, False)
        self.assertEqual(myfat.curve.get_stress(2e6), 1e9)
        self.assertEqual(myfat.curve.m, 3)

        # Test S_intercept
        myfat = FatigueParams(ultimate_stress=1e10, slope=4)
        self.assertEqual(myfat.load2stress, 1.0)
        self.assertEqual(myfat.S_ult, 1e10)
        self.assertEqual(myfat.bins, 100.0)
        self.assertEqual(myfat.goodman, False)
        self.assertEqual(myfat.curve.get_stress(1), 1e10)
        self.assertEqual(myfat.curve.m, 4)

        myfat = FatigueParams(ultimate_stress=1e10, slope=4, S_intercept=1e12)
        self.assertEqual(myfat.load2stress, 1.0)
        self.assertEqual(myfat.S_ult, 1e10)
        self.assertEqual(myfat.bins, 100.0)
        self.assertEqual(myfat.goodman, False)
        self.assertEqual(myfat.curve.get_stress(1), 1e12)
        self.assertEqual(myfat.curve.m, 4)

        # Test priority of curve setting
        myfat = FatigueParams(dnv_type='Air', dnv_name='d', Sc=1e9, Nc=2e6, slope=3, ultimate_stress=1e10, S_intercept=1e12)
        self.assertEqual(myfat.load2stress, 1.0)
        self.assertEqual(myfat.S_ult, 1e10)
        self.assertEqual(myfat.bins, 100.0)
        self.assertEqual(myfat.goodman, False)
        self.assertEqual(myfat.curve.m1, 3.0)
        self.assertEqual(myfat.curve.m2, 5.0)

        myfat = FatigueParams(Sc=1e9, Nc=2e6, slope=3, ultimate_stress=1e10, S_intercept=1e12)
        self.assertEqual(myfat.load2stress, 1.0)
        self.assertEqual(myfat.S_ult, 1e10)
        self.assertEqual(myfat.bins, 100.0)
        self.assertEqual(myfat.goodman, False)
        self.assertEqual(myfat.curve.get_stress(2e6), 1e9)
        self.assertEqual(myfat.curve.m, 3)

    def test_curve_values(self):
        Nc = 1e7
        myfat = FatigueParams(dnv_type='Air', dnv_name='b1')
        Sc = myfat.curve.get_stress(Nc)
        #self.assertAlmostEqual(Sc*1e-6, 106.97)
        myfat = FatigueParams(dnv_type='Air', dnv_name='c1')
        Sc = myfat.curve.get_stress(Nc)
        #self.assertAlmostEqual(Sc*1e-6, 65.5)
        myfat = FatigueParams(dnv_type='Air', dnv_name='d')
        Sc = myfat.curve.get_stress(Nc)
        self.assertAlmostEqual(Sc*1e-6, 52.63, 1)

        myfat = FatigueParams(dnv_type='cathodic', dnv_name='b1')
        Sc = myfat.curve.get_stress(Nc)
        #self.assertAlmostEqual(Sc*1e-6, 106.97)
        myfat = FatigueParams(dnv_type='cathodic', dnv_name='c1')
        Sc = myfat.curve.get_stress(Nc)
        #self.assertAlmostEqual(Sc*1e-6, 65.5)
        myfat = FatigueParams(dnv_type='cathodic', dnv_name='d')
        Sc = myfat.curve.get_stress(Nc)
        self.assertAlmostEqual(Sc*1e-6, 52.63, 1)
        
    def test_plotting(self):
        nn = 2*10**np.arange(10)
        myfat = FatigueParams(Sc=1e4, Nc=2e6, slope=3)
        ss = myfat.get_stress(nn)
        truth = np.logspace(6,3,nn.size)
        npt.assert_almost_equal(ss, truth)
        

    def test_rainflows(self):
        myout = read(os.path.join(DATA, FOUT[2]))
        myfat = FatigueParams(Sc=1e4, Nc=2e6, slope=3, ultimate_stress=1e7)
        N, S = myfat.get_rainflow_counts(myout['TwrBsFyt'], 50)
        self.assertEqual(N.size, 50)
        self.assertEqual(S.size, 50)

        
    def test_rainflow_markov(self):
        nbin = 20
        myout = read(os.path.join(DATA, FOUT[2]))
        myfat = FatigueParams(Sc=1e4, Nc=2e6, slope=3, ultimate_stress=1e7)
        Z, X, Y = myfat.get_rainflow_matrix(myout['TwrBsFyt'], nbin)
        self.assertEqual(X.shape, (nbin, nbin))
        self.assertEqual(Y.shape, (nbin, nbin))
        self.assertEqual(Z.shape, (nbin-1, nbin-1))

        
    def test_dels(self):
        myparam = FatigueParams(load2stress = 25.0,
                                slope = 3.0,
                                ultimate_stress = 6e8,
                                S_intercept = 5e9,
                                goodman_correction = False,
                                )
        t = np.linspace(0, 600, 10000)
        y0 = 80e3 * np.sin(2*np.pi*t/60.0)
        y80 = y0 + 80e3
        zeros = np.zeros(y0.shape)
        mydata = {"Time":t,
                  "Signal0":y0,
                  "Signal80":y80,
                  "Zeros":zeros}

        mymagnitudes = {"Mag0":["Signal0", "Zeros"],
                        "Mag80":["Signal80", "Zeros"]}

        myfatigues = {"Signal0":myparam,
                      "Signal80":myparam,
                      "Mag0":myparam,
                      "Mag80":myparam}

        myobj = AeroelasticOutput(mydata, magnitude_channels=mymagnitudes,
                                  fatigue_channels=myfatigues)

        # Test rainflow counts here too
        N, S = myparam.get_rainflow_counts(myobj['Signal80'], 50)
        idx = np.nonzero(N)[0]
        self.assertEqual(N.size, 50)
        self.assertEqual(S.size, 50)
        self.assertEqual(idx.size, 1)
        self.assertEqual(N[idx], 10)
        npt.assert_almost_equal(S[idx], 2*80e3, 2)
        
        dels, dams = myobj.get_DELs()


        self.assertAlmostEqual(dels['Signal0'], dels['Signal80'])
        self.assertGreater(dels['Signal0'], dels['Mag0'])
        self.assertAlmostEqual(dels['Signal0'], dels['Mag80'])
        
        self.assertAlmostEqual(dams['Signal0'], dams['Signal80'])
        self.assertGreater(dams['Signal0'], dams['Mag0'])
        self.assertAlmostEqual(dams['Signal0'], dams['Mag80'])
        
        dels2, dams2 = myobj.get_DELs(goodman_correction=True)
        
        self.assertGreater(dels2['Signal80'], dels2['Signal0'])
        self.assertGreater(dams2['Signal80'], dams2['Signal0'])
        
if __name__ == "__main__":
    unittest.main()



import os
import unittest
import numpy.testing as npt

from pCrunch import OpenFASTAscii, OpenFASTBinary, read, load_FAST_out

DIR = os.path.split(__file__)[0]
DATA = os.path.join(DIR, "data")

FOUT  = ['AOC_WSt.out', 'DLC2.3_1.out', 'DLC2.3_2.out', 'DLC2.3_3.out']
FOUTB = ['AOC_WSt.outb', 'Test1.outb', 'Test2.outb', 'Test3.outb', 'step_0.outb']


class Test_OpenFAST_readers(unittest.TestCase):
    
    def testOFAscii(self):
        for k, kf in enumerate(FOUT):
            fname = os.path.join(DATA, kf)
            with self.subTest(f"Reading: {kf}", i=k):
                output = OpenFASTAscii(fname)
                output.read()
                self.assertGreater(output.data.shape[0], 1)
                self.assertGreater(output.data.shape[1], 1)
                self.assertGreater(output.num_channels, 1)
    
    def testOFBinary(self):
        for k, kf in enumerate(FOUTB):
            fname = os.path.join(DATA, kf)
            with self.subTest(f"Reading: {kf}", i=k):
                output = OpenFASTBinary(fname)
                output.read()
                self.assertGreater(output.data.shape[0], 1)
                self.assertGreater(output.data.shape[1], 1)
                self.assertGreater(output.num_channels, 1)
    
    def testOFReader(self):
        for k, kf in enumerate(FOUT+FOUTB):
            fname = os.path.join(DATA, kf)
            with self.subTest(f"Reading: {kf}", i=k):
                output = read(fname)
                self.assertGreater(output.data.shape[0], 1)
                self.assertGreater(output.data.shape[1], 1)
                self.assertGreater(output.num_channels, 1)
    
    def testBatchReader(self):
        myfiles = [os.path.join(DATA, m) for m in FOUT+FOUTB]
        outputs = load_FAST_out(myfiles)
        for k, kf in enumerate(outputs):
            with self.subTest(f"Reading, {kf.filename}", i=k):
                self.assertGreater(kf.data.shape[0], 1)
                self.assertGreater(kf.data.shape[1], 1)
                self.assertGreater(kf.num_channels, 1)
    
    def testSame(self):
        outputA = read(os.path.join(DATA, FOUT[0]))
        outputB = read(os.path.join(DATA, FOUTB[0]))
        npt.assert_allclose(outputA.data, outputB.data, rtol=1e-3)


if __name__ == "__main__":
    unittest.main()



import os
import unittest
import numpy as np
import numpy.testing as npt
import pandas as pd
import pandas.testing as pdt

from pCrunch import AeroelasticOutput, FatigueParams, Crunch, read

DIR = os.path.split(__file__)[0]
DATA = os.path.join(DIR, "data")

FOUT  = ['AOC_WSt.out', 'DLC2.3_1.out', 'DLC2.3_2.out', 'DLC2.3_3.out']
FOUTB = ['AOC_WSt.outb', 'Test1.outb', 'Test2.outb', 'Test3.outb', 'step_0.outb']
FOUT1p1 = [f'DLC1.1_0_NREL5MW_OC3_spar_{m}.outb' for m in range(5)]
data = {
    "Time": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
    "WindVxi": [7, 7, 7, 7, 7, 8, 8, 8, 8, 8],
    "WindVyi": [0] * 10,
    "WindVzi": [0] * 10,
}
mc = {"Wind": ["WindVxi", "WindVyi", "WindVzi"]}
mc2 = {
    "RootMc1": ["RootMxc1", "RootMyc1", "RootMzc1"],
    "RootMc2": ["RootMxc2", "RootMyc2", "RootMzc2"],
    "RootMc3": ["RootMxc3", "RootMyc3", "RootMzc3"],
}

fc = {
    "RootMc1": FatigueParams(slope=10.0, ultimate_stress=6e8),
    "RootMc2": FatigueParams(slope=10.0, ultimate_stress=6e8),
    "RootMc3": FatigueParams(slope=10.0, ultimate_stress=6e8),
}

# Channels to focus on for extreme event tabulation
ec = [
    "RotSpeed",
    "RotThrust",
    "RotTorq",
    "RootMc1",
    "RootMc2",
    "RootMc3",
]

class Test_Crunch(unittest.TestCase):
    
    def testConstructor(self):
        # Empty
        myobj = Crunch()
        self.assertEqual(myobj.lean_flag, False)
        self.assertEqual(myobj.outputs, [])
        self.assertEqual(myobj.noutputs, 0)
        self.assertEqual(myobj.td, ())
        self.assertEqual(myobj.mc, None)
        self.assertEqual(myobj.ec, [])
        self.assertEqual(myobj.fc, {})
        npt.assert_equal(myobj.prob, np.array([]))
        pdt.assert_frame_equal(myobj.summary_stats, pd.DataFrame())
        self.assertEqual(myobj.extreme_stat, "max")
        self.assertEqual(myobj.extremes, {})
        pdt.assert_frame_equal(myobj.dels, pd.DataFrame())
        pdt.assert_frame_equal(myobj.damage, pd.DataFrame())

        # Just one
        myout = AeroelasticOutput(data)
        myobj = Crunch(myout, magnitude_channels=mc)
        self.assertEqual(myobj.lean_flag, False)
        self.assertEqual(myobj.noutputs, 1)
        self.assertEqual(myobj.td, ())
        self.assertEqual(myobj.mc, mc)
        self.assertEqual(myobj.ec, [])
        self.assertEqual(myobj.fc, {})
        npt.assert_equal(myobj.prob, [1.0])
        pdt.assert_frame_equal(myobj.summary_stats, pd.DataFrame())
        self.assertEqual(myobj.extreme_stat, "max")
        self.assertEqual(myobj.extremes, {})
        pdt.assert_frame_equal(myobj.dels, pd.DataFrame())
        pdt.assert_frame_equal(myobj.damage, pd.DataFrame())

        # Multiple
        myobj = Crunch([myout, myout], magnitude_channels=mc)
        self.assertEqual(myobj.lean_flag, False)
        self.assertEqual(myobj.noutputs, 2)
        self.assertEqual(myobj.td, ())
        self.assertEqual(myobj.mc, mc)
        self.assertEqual(myobj.ec, [])
        self.assertEqual(myobj.fc, {})
        npt.assert_equal(myobj.prob, [0.5, 0.5])
        pdt.assert_frame_equal(myobj.summary_stats, pd.DataFrame())
        self.assertEqual(myobj.extreme_stat, "max")
        self.assertEqual(myobj.extremes, {})
        pdt.assert_frame_equal(myobj.dels, pd.DataFrame())
        pdt.assert_frame_equal(myobj.damage, pd.DataFrame())

        # With trimming
        myobj = Crunch([myout, myout], magnitude_channels=mc, trim_data=[3,6], extreme_stat='abs')
        self.assertEqual(myobj.lean_flag, False)
        self.assertEqual(myobj.noutputs, 2)
        self.assertEqual(myobj.td, [3,6])
        self.assertEqual(myobj.mc, mc)
        self.assertEqual(myobj.ec, [])
        self.assertEqual(myobj.fc, {})
        npt.assert_equal(myobj.prob, [0.5, 0.5])
        pdt.assert_frame_equal(myobj.summary_stats, pd.DataFrame())
        self.assertEqual(myobj.extreme_stat, "abs")
        self.assertEqual(myobj.extremes, {})
        pdt.assert_frame_equal(myobj.dels, pd.DataFrame())
        pdt.assert_frame_equal(myobj.damage, pd.DataFrame())
        self.assertEqual(myobj.outputs[0].data.shape, (4,5))
        self.assertEqual(myobj.outputs[1].data.shape, (4,5))
        npt.assert_equal(myobj.outputs[0].time, np.arange(3,7))
        npt.assert_equal(myobj.outputs[1].time, np.arange(3,7))

        
    def testProcessOutputs(self):
        nrep = len(FOUTB)
        myouts = [read(os.path.join(DATA, m)) for m in FOUTB]
        
        # Serial
        myobj = Crunch(myouts, magnitude_channels=mc)
        myobj.process_outputs()
        self.assertEqual(myobj.noutputs, nrep)
        self.assertEqual(len(myobj.outputs), nrep)
        ss1, ext1, del1, dam1 = myobj.summary_stats, myobj.extremes, myobj.dels, myobj.damage

        # Parallel
        myobj = Crunch(myouts, magnitude_channels=mc)
        myobj.process_outputs(cores=2)
        self.assertEqual(myobj.noutputs, nrep)
        self.assertEqual(len(myobj.outputs), nrep)
        ss2, ext2, del2, dam2 = myobj.summary_stats, myobj.extremes, myobj.dels, myobj.damage

        # Lean
        myobj = Crunch(myouts, magnitude_channels=mc, lean=True)
        myobj.process_outputs()
        self.assertEqual(myobj.noutputs, nrep)
        self.assertEqual(len(myobj.outputs), 0)
        ss3, ext3, del3, dam3 = myobj.summary_stats, myobj.extremes, myobj.dels, myobj.damage

        # All the fixins
        myobj = Crunch(myouts[1:-1], magnitude_channels=mc, trim_data=(60., 500.),
                       fatigue_channels=fc, extreme_channels=ec)
        myobj.process_outputs()
        self.assertEqual(myobj.noutputs, 3)
        self.assertEqual(len(myobj.outputs), 3)
        #ss4, ext4, del4, dam4 = myobj.summary_stats, myobj.extremes, myobj.dels, myobj.damage

        pdt.assert_frame_equal(ss1, ss2)
        pdt.assert_frame_equal(ss1, ss3)

        #self.maxDiff = None
        #self.assertDictEqual(ext1, ext2)
        #self.assertDictEqual(ext1, ext3)
        
        pdt.assert_frame_equal(del1, del2)
        pdt.assert_frame_equal(del1, del3)
        
        pdt.assert_frame_equal(dam1, dam2)
        pdt.assert_frame_equal(dam1, dam3)

    def testStreamingByOutput(self):
        nrep       = len(FOUTB)
        myouts     = [read(os.path.join(DATA, m)) for m in FOUTB]
        myobj_base = Crunch(myouts, magnitude_channels=mc)
        myobj_base.process_outputs()
        myobj      = Crunch(lean=True, magnitude_channels=mc)

        nrep = len(FOUTB)
        for k in range(nrep):
            myobj.add_output(myouts[k])

        self.assertEqual(myobj_base.noutputs, nrep)
        self.assertEqual(myobj.noutputs, nrep)
        self.assertEqual(len(myobj_base.outputs), nrep)
        self.assertEqual(len(myobj.outputs), 0)

        pdt.assert_frame_equal(myobj_base.summary_stats, myobj.summary_stats, check_like=True)
        pdt.assert_frame_equal(myobj_base.dels, myobj.dels, check_column_type=False)
        pdt.assert_frame_equal(myobj_base.damage, myobj.damage, check_column_type=False)

    def testStreamingByStats(self):
        nrep       = len(FOUTB)
        myouts     = [read(os.path.join(DATA, m)) for m in FOUTB]
        myobj_base = Crunch(myouts, magnitude_channels=mc)
        myobj_base.process_outputs()
        myobj      = Crunch(lean=True, magnitude_channels=mc)

        nrep = len(FOUTB)
        for k in range(nrep):
            fname, stats, extremes, dels, damage =  myobj.process_single(myouts[k])
            myobj.add_output_stats(fname, stats, extremes, dels, damage)

        self.assertEqual(myobj_base.noutputs, nrep)
        self.assertEqual(myobj.noutputs, nrep)
        self.assertEqual(len(myobj_base.outputs), nrep)
        self.assertEqual(len(myobj.outputs), 0)

        pdt.assert_frame_equal(myobj_base.summary_stats, myobj.summary_stats, check_like=True)
        pdt.assert_frame_equal(myobj_base.dels, myobj.dels, check_column_type=False)
        pdt.assert_frame_equal(myobj_base.damage, myobj.damage, check_column_type=False)

    def testDels(self):
        self.assertTrue(True)

    def testLoadRankings(self):
        myouts = [read(os.path.join(DATA, m)) for m in FOUTB]
        
        # Serial
        myobj = Crunch(myouts, magnitude_channels=mc)
        myobj.process_outputs()
        print(myobj.get_load_rankings(['RotTorq','GenPwr'], ['mean','max']))
        self.assertTrue(True)

    def testWindspeeds(self):
        myout = AeroelasticOutput(data)
        myouts = [myout]*10
        
        # Read the output
        myobj = Crunch(myouts, magnitude_channels=mc)
        windout = myobj._get_windspeeds('Wind')
        npt.assert_equal(windout, 7.5*np.ones(10))
        windout = myobj._get_windspeeds('Wind', idx=[1,4])
        npt.assert_equal(windout, 7.5*np.ones(2))

        # Read the stats log
        myobj = Crunch(myouts, magnitude_channels=mc, lean=True)
        myobj.process_outputs()
        windout = myobj._get_windspeeds('Wind')
        npt.assert_equal(windout, 7.5*np.ones(10))
        windout = myobj._get_windspeeds('Wind', idx=[1,4])
        npt.assert_equal(windout, 7.5*np.ones(2))

        # Provide the input
        myobj = Crunch([myout]*5, magnitude_channels=mc, lean=True)
        myobj.process_outputs()
        windout = myobj._get_windspeeds([1,2,3,4,5])
        npt.assert_equal(windout, [1,2,3,4,5])
        windout = myobj._get_windspeeds([1,2,3,4,5], idx=[1,4])
        npt.assert_equal(windout, [2,5])
        windout = myobj._get_windspeeds([4,5], idx=[1,4])
        npt.assert_equal(windout, [4,5])

        # Expect error
        with self.assertRaises(ValueError):
            windout = myobj._get_windspeeds([1,2,3,4])
        with self.assertRaises(ValueError):
            windout = myobj._get_windspeeds([1,2,3,4], idx=[1,4])

    def test_prob(self):
        myout = AeroelasticOutput(data)
        myouts = [myout]*10
        
        myobj = Crunch(myouts, magnitude_channels=mc)
        myobj.process_outputs()
        
        myobj.set_probability_wind_distribution('Wind', 7.5, kind='weibull')
        npt.assert_equal(myobj.prob, 0.1*np.ones(10))
        
        myobj.set_probability_wind_distribution('Wind', 7.5, kind='weibull', idx=[1,4])
        npt.assert_equal(myobj.prob[[1,4]], 0.5*np.ones(2))
        
        myobj.set_probability_wind_distribution('Wind', 7.5, kind='rayleigh')
        npt.assert_equal(myobj.prob, 0.1*np.ones(10))
        
        myobj.set_probability_wind_distribution('Wind', 7.5, kind='rayleigh', idx=[1,4])
        npt.assert_equal(myobj.prob[[1,4]], 0.5*np.ones(2))
        
        myobj.set_probability_wind_distribution('Wind', 7.5, kind='uniform')
        npt.assert_equal(myobj.prob, 0.1*np.ones(10))
        
        myobj.set_probability_wind_distribution('Wind', 7.5, kind='uniform', idx=[1,4])
        npt.assert_equal(myobj.prob[[1,4]], 0.5*np.ones(2))
        
        myobj.set_probability_wind_distribution('Wind', 7.5, kind='bleh')
        npt.assert_equal(myobj.prob, 0.1*np.ones(10))
        
        myobj.set_probability_wind_distribution('Wind', 7.5, kind='bleh', idx=[1,4])
        npt.assert_equal(myobj.prob[[1,4]], 0.5*np.ones(2))
        
        myobj.set_probability_wind_distribution(np.r_[15*np.ones(9), 7.5], 7.5, kind='weibull')
        npt.assert_equal(myobj.prob[:-1], myobj.prob[0])
        self.assertGreater(myobj.prob[-1], myobj.prob[-2])

        myobj.set_probability_turbine_class('Wind', 1)
        myobj.set_probability_turbine_class('Wind', 2)
        myobj.set_probability_turbine_class('Wind', 3)
        myobj.set_probability_turbine_class('Wind', 1, idx=[1,4])

    def test_aep(self):
        myout = AeroelasticOutput(data)
        myouts = [myout]*10

        aep_true = myout.compute_energy('WindVxi') * 8760/myout.elapsed_time
        
        myobj = Crunch(myouts, magnitude_channels=mc)
        aep_w1, aep_uw1 = myobj.compute_aep('Wind')
        self.assertAlmostEqual(aep_true, aep_w1, 5)
        self.assertAlmostEqual(aep_w1, aep_uw1, 5)
        
        myobj = Crunch(myouts, magnitude_channels=mc, lean=True)
        myobj.process_outputs()
        aep_w2, aep_uw2 = myobj.compute_aep('Wind')
        self.assertAlmostEqual(aep_w1, aep_w2, 5)
        self.assertAlmostEqual(aep_w2, aep_uw2, 5)
        
        aep_w3, aep_uw3 = myobj.compute_aep('Wind', loss_factor=0.5)
        self.assertAlmostEqual(0.5*aep_w1, aep_w3, 5)
        self.assertAlmostEqual(aep_w3, aep_uw3, 5)
        
        aep_w4, aep_uw4 = myobj.compute_aep('Wind', idx=[1,4])
        self.assertAlmostEqual(aep_w1, aep_w4, 5)
        self.assertAlmostEqual(aep_w4, aep_uw4, 5)
        
        myobj.prob[0] = 0.0
        aep_w5, aep_uw5 = myobj.compute_aep('Wind')
        self.assertAlmostEqual(aep_w1, aep_w5, 5)
        self.assertAlmostEqual(aep_w4, aep_uw4, 5)

        
    def test_total_fatigue(self):
        # Test bulk vs stream
        myouts = [read(os.path.join(DATA, 'DLC1p1', m)) for m in FOUT1p1]
        myobj = Crunch(myouts, magnitude_channels=mc2, fatigue_channels=fc)
        myobj.process_outputs(compute_damage = True)
        del1, dam1 = myobj.compute_total_fatigue()

        myobj2 = Crunch(myouts, lean=True, magnitude_channels=mc2, fatigue_channels=fc)
        myobj2.process_outputs(compute_damage = True)
        del2, dam2 = myobj2.compute_total_fatigue()

        pdt.assert_frame_equal(del1, del2)
        pdt.assert_frame_equal(dam1, dam2)

        # Test prob weighting and index
        myobj.prob[-1] = 0.0
        del3, dam3 = myobj.compute_total_fatigue()
        myobj.prob[-1] = myobj.prob[-2]
        del4, dam4 = myobj.compute_total_fatigue(idx=[0,1,2,3])
        pdt.assert_series_equal(del3.loc['Weighted'], del4.loc['Weighted'])
        pdt.assert_series_equal(dam3.loc['Weighted'], dam4.loc['Weighted'])

        # Test lifetime and availability- only for damage now
        myobj = Crunch(myouts, magnitude_channels=mc2, fatigue_channels=fc)
        myobj.process_outputs(compute_damage = True)
        del5, dam5 = myobj.compute_total_fatigue(lifetime=30, availability=1.0)
        del6, dam6 = myobj.compute_total_fatigue(lifetime=30, availability=0.75)
        del7, dam7 = myobj.compute_total_fatigue(lifetime=15, availability=1.0)
        pdt.assert_frame_equal(0.75*dam5, dam6)
        pdt.assert_frame_equal(0.5*dam5, dam7)

        # Test faults
        del8, dam8 = myobj.compute_total_fatigue(lifetime=30, availability=1.0, idx_fault=[2], n_fault=3)
        pdt.assert_series_equal(3*myobj.damage.iloc[2], dam8.loc['Weighted'], check_names=False)
        pdt.assert_series_equal(3*myobj.damage.iloc[2], dam8.loc['Unweighted'], check_names=False)
        
        # Test index conflicts
        try:
            _, _ = myobj.compute_total_fatigue(lifetime=15, availability=0.75, idx=[0,1,2], idx_park=[2])
        except ValueError:
            self.assertTrue(True)
        try:
            _, _ = myobj.compute_total_fatigue(lifetime=15, availability=0.75, idx=[0,1,2], idx_fault=[2])
        except ValueError:
            self.assertTrue(True)
        try:
            _, _ = myobj.compute_total_fatigue(lifetime=15, availability=0.75, idx_park=[0,1,2], idx_fault=[2])
        except ValueError:
            self.assertTrue(True)
            
        
    def testProperties(self):
        myout = AeroelasticOutput(data)
        myouts = [myout]*10
        
        myobj = Crunch(myouts, magnitude_channels=mc)
        npt.assert_equal(myobj.num_timesteps(), [10]*10)
        npt.assert_equal(myobj.elapsed_time(), [9]*10)
        npt.assert_equal(myobj.num_channels(), [5]*10)
        npt.assert_equal(myobj.idxmins(), [np.zeros(5)]*10)
        npt.assert_equal(myobj.idxmaxs(), [[9, 5, 0, 0, 5]]*10)
        npt.assert_equal(myobj.minima(), [[1, 7, 0, 0, 7]]*10)
        npt.assert_equal(myobj.maxima(), [[10, 8, 0, 0, 8]]*10)
        npt.assert_equal(myobj.absmaxima(), [[10, 8, 0, 0, 8]]*10)
        npt.assert_equal(myobj.ranges(), [[9, 1, 0, 0, 1]]*10)
        npt.assert_equal(myobj.variable(), [[0, 1, 4]]*10)
        npt.assert_equal(myobj.constant(), [[2, 3]]*10)
        npt.assert_equal(np.array(myobj.sums()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.sums_squared()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.sums_cubed()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.sums_fourth()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.second_moments()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.third_moments()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.fourth_moments()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.means()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.medians()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.stddevs()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.skews()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.kurtosis()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.integrated()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.compute_energy('WindVxi')), myobj.integrated()[0][1])
        npt.assert_equal(np.array(myobj.total_travel('WindVxi')), np.ones(10))

        # Can only do some stats without the output list
        myobj = Crunch(myouts, magnitude_channels=mc, lean=True)
        myobj.process_outputs()
        #npt.assert_equal(myobj.num_timesteps(), [10]*10)
        npt.assert_equal(myobj.elapsed_time(), [9]*10)
        #npt.assert_equal(myobj.num_channels(), [5]*10)
        #npt.assert_equal(myobj.idxmins(), [np.zeros(5)]*10)
        #npt.assert_equal(myobj.idxmaxs(), [[9, 5, 0, 0, 5]]*10)
        npt.assert_equal(myobj.minima(), [[1, 7, 0, 0, 7]]*10)
        npt.assert_equal(myobj.maxima(), [[10, 8, 0, 0, 8]]*10)
        #npt.assert_equal(myobj.absmaxima(), [[10, 8, 0, 0, 8]]*10)
        npt.assert_equal(myobj.ranges(), [[9, 1, 0, 0, 1]]*10)
        #npt.assert_equal(myobj.variable(), [[0, 1, 4]]*10)
        #npt.assert_equal(myobj.constant(), [[2, 3]]*10)
        #npt.assert_equal(np.array(myobj.sums()).shape, (10,myobj.num_channels()[0]))
        #npt.assert_equal(np.array(myobj.sums_squared()).shape, (10,myobj.num_channels()[0]))
        #npt.assert_equal(np.array(myobj.sums_cubed()).shape, (10,myobj.num_channels()[0]))
        #npt.assert_equal(np.array(myobj.sums_fourth()).shape, (10,myobj.num_channels()[0]))
        #npt.assert_equal(np.array(myobj.second_moments()).shape, (10,myobj.num_channels()[0]))
        #npt.assert_equal(np.array(myobj.third_moments()).shape, (10,myobj.num_channels()[0]))
        #npt.assert_equal(np.array(myobj.fourth_moments()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.means()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.medians()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.stddevs()).shape, (10,myobj.num_channels()[0]))
        #npt.assert_equal(np.array(myobj.skews()).shape, (10,myobj.num_channels()[0]))
        #npt.assert_equal(np.array(myobj.kurtosis()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.integrated()).shape, (10,myobj.num_channels()[0]))
        npt.assert_equal(np.array(myobj.compute_energy('WindVxi')), myobj.integrated()[0][1])
    
if __name__ == "__main__":
    unittest.main()
    



# Copyright 2021 NREL

# Licensed under the Apache License, Version 2.0 (the "License"); you may not use
# this file except in compliance with the License. You may obtain a copy of the
# License at http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software distributed
# under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
# CONDITIONS OF ANY KIND, either express or implied. See the License for the
# specific language governing permissions and limitations under the License.

import setuptools

setuptools.setup()



import multiprocessing as mp
from functools import partial
import numpy as np
import pandas as pd
from .utility import weibull_mean, rayleigh_mean

    
class Crunch:
    """Aeroelastic time-series simulation batch postprocessing class."""

    def __init__(self, outputs=[], **kwargs):
        """
        Creates an instance of `Crunch`.

        Parameters
        ----------
        outputs : list (optional)
            List of OpenFAST output filepaths or dicts of OpenFAST outputs.
        lean : boolean (optional)
            If False (default), the outputs are kept in memory as they are processed.  If True, only summary statistics are retained.
        fatigue_channels : dict (optional)
            Dictionary with format:
            'channel': 'fatigue slope'
        extreme_stat : str (optional)
            Whether the extreme event calculation should work on [max, min, abs].
            Default, 'max'
        extreme_channels : list (optional)
            Limit calculation of extremes to the channel names in the list.  Unspecified means all channels are processed and reported.
        magnitude_channels : dict (optional)
            Additional channels as vector magnitude of other channels.
            Format: 'new-chan': ['chan1', 'chan2', 'chan3']
        trim_data : tuple or list (optional)
            Trim processed outputs to desired times.
            Format: (min, max)
        probability : list, tuple, or Numpy array (optional)
            Probability or weighting for each output time series.
            Should be a list or tuple or array of floats the same length as the `outputs` list
        """

        self.lean_flag = kwargs.get("lean", False)
        
        if isinstance(outputs, list):
            self.outputs = outputs
        else:
            self.outputs = [outputs]
        self.noutputs = len(self.outputs)

        self.td = kwargs.get("trim_data", ())
        self.trim_data(*self.td)

        self.mc = kwargs.get("magnitude_channels", None)
        self.append_magnitude_channels(self.mc)
        
        self.ec = kwargs.get("extreme_channels", [])
        self.extreme_stat = kwargs.get("extreme_stat", "max")
        
        self.fc = kwargs.get("fatigue_channels", {})
        for k in range(self.noutputs):
            self.outputs[k].fc = self.fc
            self.outputs[k].ec = self.ec
        
        self.prob = kwargs.get("probability", np.array([]))
        if self.prob.size == 0:
            self._reset_probabilities()
        
        # Initialize containers
        self.summary_stats = pd.DataFrame()
        self.extremes = {}
        self.dels = pd.DataFrame()
        self.damage = pd.DataFrame()

    def copy(self):
        mycp =  type(self)(outputs = self.outputs,
                           lean = self.lean_flag,
                           trim_data = self.td,
                           magnitude_channels = self.mc,
                           extreme_channels = self.ec,
                           fatigue_channels = self.fc,
                           probability = self.prob)
        mycp.summary_stats = self.summary_stats.copy()
        mycp.extremes = self.extremes.copy()
        mycp.dels = self.dels.copy()
        mycp.damage = self.damage.copy()
        return mycp
        
    def process_outputs(self, cores=1, **kwargs):
        """
        Processes all outputs for summary statistics and configured damage
        equivalent loads.
        """
        
        if cores > 1:
            pool = mp.Pool(processes=cores)
            returned = pool.map(
                partial(self.process_single, **kwargs), self.outputs
            )
            pool.close()
            pool.join()

        else:
            returned = [self.process_single(output, **kwargs) for output in self.outputs]

        stats  = {}
        extrs  = {}
        dels   = {}
        damage = {}

        for ifile, istats, iextrs, idels, idamage in returned:
            if ifile == '':
                ifile = len(stats)
            stats[ifile]  = istats
            extrs[ifile]  = iextrs
            dels[ifile]   = idels
            damage[ifile] = idamage
            
        self.summary_stats, self.extremes, self.dels, self.damage = self.post_process(stats, extrs, dels, damage)
        
        if self.lean_flag:
            self.outputs = []
        

    def process_single(self, output, **kwargs):
        """
        Process AeroelasticOutput output `f`.

        Parameters
        ----------
        f : str | AerolelasticOutput
            Path to output or direct output in dict format.
        """
        
        if not hasattr(output, 'stats'):
            output.trim_data(*self.td)
            output.append_magnitude_channels(self.mc)
            output.fc.update(self.fc)
            output.ec = list(set(output.ec + self.ec))
            output.set_extreme_stat(self.extreme_stat)
            output.process(**kwargs)

        return output.filename, output.stats, output.ext_table, output.dels, output.damage


    def add_output(self, output, **kwargs):
        """
        Appends output to the list and processes it

        Parameters
        ----------
        output : AerolelasticOutput
        """
        
        # Analyze the data
        fname, stats, extremes, dels, damage =  self.process_single(output, **kwargs)

        # Add to output stat containers
        self.add_output_stats(fname, stats, extremes, dels, damage)

        # Add the output to our lists
        if not self.lean_flag:
            self.outputs.append(output)

        
    def add_output_stats(self, fname, stats, extremes, dels, damage):

        # Append to the stats logs
        statsdf = self._process_summary_stats( {fname:stats} )
        ndf = len(statsdf.index)+len(self.summary_stats.index)
        nidx = len(set(list(statsdf.index)+list(self.summary_stats.index)))
        ignore_flag = nidx < ndf
        self.summary_stats = pd.concat((self.summary_stats, statsdf), axis=0,
                                       ignore_index=ignore_flag)

        # Extreme events- join dictionaries
        self.extremes = self._process_extremes( {'null':self.extremes, fname:extremes} )
        
        # Damage and Damage Equivalent Loads
        self.dels   = pd.concat((self.dels,   pd.DataFrame(dels, index=[fname])), axis=0)
        self.damage = pd.concat((self.damage, pd.DataFrame(damage, index=[fname])), axis=0)

        # Increment counters
        self.noutputs += 1
        self._reset_probabilities()


    def _process_summary_stats(self, statsin):
        # Summary statistics
        ss = pd.DataFrame.from_dict(statsin, orient="index").stack().to_frame()
        ss = pd.DataFrame(ss[0].values.tolist(), index=ss.index)
        return ss.unstack().swaplevel(axis=1)

    def _process_extremes(self, extremesin):
        # Extremes
        extreme_table = {}
        for _, d in extremesin.items():
            for channel, sub in d.items():
                subt = sub if isinstance(sub, list) else [sub]
                if channel not in extreme_table:
                    extreme_table[channel] = subt
                else:
                    extreme_table[channel] += subt
        return extreme_table
        
    def post_process(self, statsin, extremesin, delsin, damagein):
        """Post processes internal data to produce DataFrame outputs."""
        summary_stats = self._process_summary_stats(statsin)

        # Extreme events
        extremes = self._process_extremes(extremesin)
        
        # Damage and Damage Equivalent Loads
        dels = pd.DataFrame(delsin).T
        damage = pd.DataFrame(damagein).T

        return summary_stats, extremes, dels, damage

    
    def get_load_rankings(self, ranking_vars, ranking_stats):
        """
        Returns load rankings across all outputs in `self.outputs`.

        Parameters
        ----------
        rankings_vars : list
            List of variables to evaluate for the ranking process.
        ranking_stats : list
            Summary statistic to evalulate. Currently supports 'min', 'max',
            'abs', 'mean', 'std'.
        """
        if not isinstance(ranking_vars, (list, tuple, set)):
            ranking_vars = [ranking_vars]
            #raise ValueError('Need a list or tuple of ranking variables')
        if not isinstance(ranking_stats, (list, tuple, set)):
            ranking_stats = [ranking_stats]
            #raise ValueError('Need a list or tuple of ranking statistics')

        summary_stats = self.summary_stats.copy().swaplevel(axis=1).stack(future_stack=True)

        out = []
        for var in ranking_vars:
            for stat in ranking_stats:

                col = pd.MultiIndex.from_product([self.summary_stats.index, [var]])
                if stat in ["max", "abs"]:
                    res = (*summary_stats.loc[col][stat].idxmax(), stat,
                           summary_stats.loc[col][stat].max())

                elif stat == "min":
                    res = (*summary_stats.loc[col][stat].idxmin(), stat,
                           summary_stats.loc[col][stat].min())

                elif stat in ["mean", "std"]:
                    res = ("NA", var, stat, summary_stats.loc[col][stat].mean())

                elif stat == "median":
                    res = ("NA", var, stat, summary_stats.loc[col][stat].median())

                else:
                    raise NotImplementedError(f"Statistic '{stat}' not supported for load ranking.")

                out.append(res)

        return pd.DataFrame(out, columns=["file", "channel", "stat", "val"])


    def _get_windspeeds(self, windspeed, idx=None):
        """
        Returns the wind speeds in the output list if not already known.

        Parameters
        ----------
        windspeed : str or list / Numpy array
            If input as a list or array, it is simply returned (after checking for correct length).
            If input as a string, then it should be the channel name of the u-velocity component
            ('Wind1VelX` in OpenFAST)
        idx: list or Numpy array (default None)
            Index vector into output case list
        
        Returns
        ----------
        windspeed : Numpy array
            Numpy array of the average inflow wind speed for each output.
        """

        if isinstance(windspeed, str):
            # If user passed wind u-component channel name (e.g. 'Wind1VelX'),
            # then get mean value for every case
            if len(self.summary_stats) > 0:
                mywindspeed = self.summary_stats[windspeed]['mean'].to_numpy()
            else:
                mywindspeed = np.array( [np.mean( k[windspeed] ) for k in self.outputs] )
            if idx is not None and len(idx) > 0:
                mywindspeed = mywindspeed[idx]
        else:
            # Assume user passed windspeeds directly
            mywindspeed = np.array(windspeed)
            if idx is not None and len(idx) > 0:
                ntarget = len(idx)
                if mywindspeed.size == self.noutputs:
                    mywindspeed = mywindspeed[idx]
            else:
                ntarget = self.noutputs
            if mywindspeed.size != ntarget:
                raise ValueError(f"When giving list of windspeeds, must have {ntarget} values, one for every case")
            
        return mywindspeed

    
    def _reset_probabilities(self):
        if self.noutputs > 0:
            self.prob = np.ones( self.noutputs ) / float(self.noutputs)
            
    def set_probability_wind_distribution(self, windspeed, v_avg, weibull_k=2.0, kind="weibull", method='pdf', idx=None, v_prob=None, probability=None):
        """
        Sets the probability of each output in the list based on a Weibull or Rayleigh or Uniform
        distribution for windspeed.

        For the 'pdf' method, the probability density function is sampled at each windspeed.
        For the 'cdf' method, the sorted vector of windspeeds is used to create probability bins, with the
        difference of the cumulative density function values at bin edges used to assign probability mass values.

        In all cases, the resulting vector of probability values is then scaled such that they sum to one.

        Parameters
        ----------
        windspeed : str or list / Numpy array
            If input as a list or array, it is simply returned (after checking for correct length).
            If input as a string, then it should be the channel name of the u-velocity component
            ('Wind1VelX` in OpenFAST)
        v_avg : float
            Average velocity for the wind distribution (sets the scale parameter in the distribution)
        weibull_k : float (optional)
            Shape parameter for the Weibull distribution. Defaults to 2.0
        kind : str (optional)
            Which distribution to use.  Should be either 'weibull' or 'rayleigh' or 'uniform' or 'user'.
            Default is 'weibull'
        method : str (optional)
            Which distribution to use.  Should be either 'pdf' or 'cdf'.
            Default is 'pdf'
        idx: list or Numpy array (optional)
            Index vector into output case list.  Default is None
        v_prob: list or Numpy array (optional)
            User defined wind speed for indexing probability
        probability: list or Numpy array (optional)
            User defined probability, indexed by v_prob
        """

        # Input consistency and sanity checks
        if isinstance(windspeed, (float, int)) or len(windspeed) == 1:
            method = 'pdf'
            kind = 'uniform'
        if kind.lower() == 'uniform':
            method = 'pdf'
        if method.lower() not in ['pdf', 'cdf']:
            print(f"Unknown method, {method.lower()}. Expected 'pdf' or 'cdf'. Defaulting to 'pdf'")
            method = 'pdf'
        if kind.lower() not in ['weibull', 'rayleigh', 'rayliegh', 'uniform', 'user']:
            print(f"Unknown probability distribution, {kind}, defaulting to uniform")
            kind = 'uniform'

        if kind == 'user':
            if not probability:
                raise Exception("The user must define a probability vs wind speed (v_prob)")
            if not v_prob:
                raise Exception("The user must define a v_prob vs wind speed (v_prob)")
            if len(probability) != len(v_prob):
                raise Exception(f"The length of probability ({len(probability)}) must match the length of v_prob ({len(v_prob)})")
            
        # Get windspeed from user or cases
        mywindspeed = self._get_windspeeds(windspeed, idx=idx)

        # Set probability from distributions
        if method.lower() == 'pdf':
            if kind.lower() == "weibull":
                prob = weibull_mean(mywindspeed, weibull_k, v_avg, kind="pdf")

            elif kind.lower() in ["rayleigh", "rayliegh"]:
                prob = rayleigh_mean(mywindspeed, v_avg, kind="pdf")

            elif kind.lower() == "uniform":
                prob = np.ones(mywindspeed.shape)


        elif method.lower() == 'cdf':
            # Get unique wind speeds
            wind_sort, unique2wind = np.unique(mywindspeed, return_inverse=True)

            # Create wind speed bins for the CDF
            bins = 0.5 * (wind_sort[:-1] + wind_sort[1:])
            bins = np.r_[0.0, bins, np.inf]

            # Get bin edge probabilities
            if kind.lower() == "weibull":
                prob_edge = weibull_mean(wind_sort, weibull_k, v_avg, kind="cdf")

            elif kind.lower() in ["rayleigh", "rayliegh"]:
                prob_edge = rayleigh_mean(wind_sort, v_avg, kind="cdf")

            # Get bin integral probabilities
            prob_unique = np.diff(prob_edge)

            # Put back in regular indexing
            prob = prob_unique[unique2wind]

        if kind == 'user':
            prob = np.interp(windspeed,v_prob,probability)

        # Ensure probability sums to one for all of our cases
        self._reset_probabilities()
        prob = prob / prob.sum()
        if idx is not None and len(idx) > 0:
            self.prob[idx] = prob
        else:
            self.prob = prob

        
    def set_probability_turbine_class(self, windspeed, turbine_class, idx=None):
        """
        Sets the probability of each output in the list based on a Weibull distribution
        (shape parameter 2) and average velocity as determined by IEC turbine class standards.

        Parameters
        ----------
        windspeed : str or list / Numpy array
            If input as a list or array, it is simply returned (after checking for correct length).
            If input as a string, then it should be the channel name of the u-velocity component
            ('Wind1VelX` in OpenFAST)
        turbine_class : int or str
            Turbine class either 1/'I' or 2/'II' or 3/'III'
        idx: list or Numpy array (default None)
            Index vector into output case list
        """

        # IEC average velocities
        if turbine_class in [1, "I"]:
            Vavg = 50 * 0.2

        elif turbine_class in [2, "II"]:
            Vavg = 42.5 * 0.2

        elif turbine_class in [3, "III"]:
            Vavg = 37.5 * 0.2

        else:
            raise ValueError(f"Unknown turbine class, {turbine_class}")

        self.set_probability_wind_distribution(windspeed, Vavg, weibull_k=2.0, kind="weibull", idx=idx)

        
    def compute_aep(self, pwrchan, loss_factor=0.0, idx=None):
        """
        Computes annual energy production [kWh] based on all outputs in the list.

        Parameters
        ----------
        pwrchan : string
            Name of channel containing output electrical power in the simulation
            (e.g. 'GenPwr' in OpenFAST)
        loss_factor : float
            Loss factor for availability and other losses
            (soiling, array, etc.) to apply to AEP calculation (0.0 = no losses)"
        idx: list or Numpy array (default None)
            Index vector into output case list

        Returns
        ----------
        aep_weighted_kWh : float
            Weighted AEP where each output is weighted by the probability of
            occurence determined from its average wind speed
        aep_unweighted_kWh : float
            Unweighted AEP that assumes all outputs in the list are equally
            likely (just a mean)
        """

        # Energy for every output case
        E = np.array( self.compute_energy(pwrchan) )
        T = np.array( self.elapsed_time() )
        prob = self.prob.copy()

        if idx is not None and len(idx) > 0:
            E = E[idx]
            T = T[idx]
            prob = prob[idx]

        # Make sure our probabilities sum appropriately
        prob = prob / prob.sum()

        # Calculate scaling factor for year with losses included
        fact = (1.0 - loss_factor) * 365.0 * 24.0 #* 60.0 * 60.0 # Add these in for kW-s

        # Sum with probability.  Finding probability weighted average per second, scaled by total seconds
        aep_weighted_kWh = fact * np.dot(E, prob) / np.dot(T, prob)

        # Assume equal probability
        prob = np.ones(E.shape)/E.size
        aep_unweighted_kWh = fact * np.dot(E, prob) / np.dot(T, prob)

        return aep_weighted_kWh, aep_unweighted_kWh

    
    def compute_total_fatigue(self, lifetime=0.0, availability=1.0,
                              idx=None, idx_park=None, idx_fault=None, n_fault=0):
        """
        Computes total damage equivalent load (DEL) and Palmgren-Miner total damage based on the
        outputs in the list.  For DEL, only the optional input, 'idx' is used to select the correct outputs.
        Aside from the windspeed probability, no other lifetime scaling is applied.  For the damage calculation,
        in addition to 'lifetime' scaling, there is an allowance for operational runs with 'availability',
        parked rotor simulations for downtime with 'idx_park' scaled with (1 - availability), and expected number
        of fault events in a lifetime with 'idx_fault' and 'n_fault'.
        All optional inputs are valid for the damage calculation.

        The `process_outputs` function shuld be run before this if the
        output statisctics were not added in streaming mode.

        Parameters
        ----------
        lifetime: float (optional)
            Number of years of expected service life of the turbine
        availability: float (optional)
            Fraction of time turbine is expected to be available for power production
            versus down for maintenance.  This availability should not include expected time that
            wind speed is below cutin or above cutout.
        idx: list or Numpy array (optional)
            Index vector into output case list for operational cases (even for wind speed
            less than cutin or greater than cutout)
        idx: list or Numpy array (optional)
            Index vector into output case list for operational cases (even for wind speed
            less than cutin or greater than cutout)
        idx_park: list or Numpy array (optional)
            Index vector into output case list for cases when rotor is parked due to maintenance downtime
            (as opposed to parked to due below cutin or above cutout)
        idx_fault: list or Numpy array (optional)
            Index vector into output case list for cases when rotor is parked due to maintenance downtime
            (as opposed to parked to due below cutin or above cutout).  If this is set, must also provide 'n_fault'
        n_fault: int or float (optional)
            Number of fault events in the idx_fault list that are expected to occur in the turbine liftime.
            If this is set, must also provide 'idx_fault'

        Returns
        ----------
        total_dels : Pandas DataFrame
            Weighted and unweighted summations of damage equivalent loads for each fatigue channel
        total_damage : Pandas DataFrame
            Weighted and unweighted summations of lifetime scaled damage for each fatigue channel
        """
        dels   = self.dels.fillna(0.0)
        damage = self.damage.fillna(0.0)
        dels_total = None
        damage_total = None

        # Convert none inputs
        if idx is None:
            idx = []
        if idx_park is None:
            idx_park = []
        if idx_fault is None:
            idx_fault = []
            
        # Handle generic "run everything" input
        if len(idx) == 0 and len(idx_park) == 0 and len(idx_fault)==0:
            idx = [m for m in range(self.noutputs)]

        # Look for bad index inputs
        if np.any( np.intersect1d(idx, idx_park) ):
            raise ValueError('Cannot have common indices in idx and idx_park')
        if np.any( np.intersect1d(idx, idx_fault) ):
            raise ValueError('Cannot have common indices in idx and idx_fault')
        if np.any( np.intersect1d(idx_park, idx_fault) ):
            raise ValueError('Cannot have common indices in idx_park and idx_fault')

        
        if len(dels) > 0:
            prob = self.prob.copy()
            
            if len(idx) > 0:
               dels = dels.iloc[idx]
               prob = self.prob[idx]

            # Make sure our probabilities sum appropriately
            prob = prob / prob.sum()
                
            dels_weighted = np.sum(prob[:,np.newaxis] * dels, axis=0)
            dels_unweighted = dels.mean(axis=0)
            
            dels_total = pd.DataFrame([dels_weighted, dels_unweighted],
                                      index=['Weighted','Unweighted'])
            
        if len(damage) > 0:
            # Will need to account for any differences in elapsed time
            T_full = np.array( self.elapsed_time() )

            # Lifetime scaling factor days * hours * min * sec
            factor = lifetime * 365.0*24.0*60.0*60.0 if lifetime > 0.0 else 1.0

            # Initialize outputs
            damage_weighted_life = 0.0
            damage_unweighted_life = 0.0
            
            # First do operational cases
            if len(idx) > 0:
               damage = damage.iloc[idx]
               T = T_full[idx]
               prob = self.prob[idx]
               prob = prob / prob.sum()
                
               damage_weighted = np.sum(prob[:,np.newaxis] * damage, axis=0)
               T_weighted = np.dot(T, prob)
               damage_weighted_life += availability * damage_weighted * factor / T_weighted
               
               damage_unweighted = damage.mean(axis=0)
               T_unweighted = T.mean()
               damage_unweighted_life += availability * damage_unweighted * factor / T_unweighted

            # Now add in non-operational (parked) cases
            if len(idx_park) > 0:
                damage = self.damage.fillna(0.0).iloc[idx_park]
                T = T_full[idx_park]
                prob = self.prob[idx_park]
                prob = prob / prob.sum()

                damage_weighted = np.sum(prob[:,np.newaxis] * damage, axis=0)
                T_weighted = np.dot(T, prob)
                damage_weighted_life += (1.0 - availability) * damage_weighted * factor / T_weighted

                damage_unweighted = damage.mean(axis=0)
                T_unweighted = T.mean()
                damage_unweighted_life += (1.0 - availability) * damage_unweighted * factor / T_unweighted
                
            # Now add in fault events (fault cases)
            if len(idx_fault) > 0:
                if n_fault == 0:
                    print('Warning: fault case indices provided, but n_fault=0')
                damage = n_fault * np.sum(self.damage.fillna(0.0).iloc[idx_fault], axis=0)
                damage_weighted_life += damage
                damage_unweighted_life += damage
            
            # Results in dataframe
            damage_total = pd.DataFrame([damage_weighted_life, damage_unweighted_life],
                                        index=['Weighted','Unweighted'])

        
        return dels_total, damage_total

        
    # Batch versions of AeroelasticOutput methods
    def calculate_channel(self, instr, namein):
        for k in self.outputs:
            k.calculate_channel(instr, namein)
            
    def trim_data(self, tmin=0, tmax=np.inf):
        for k in self.outputs:
            k.trim_data(tmin, tmax)
            
    def append_magnitude_channels(self, magnitude_channels=None):
        for k in self.outputs:
            k.append_magnitude_channels(magnitude_channels)
            
    def add_magnitude_channels(self, magnitude_channels=None):
        self.append_magnitude_channels(magnitude_channels)
        
    def add_load_rose(self, load_rose=None, nsec=6):
        for k in self.outputs:
            k.add_load_rose(load_rose=load_rose, nsec=nsec)

    def drop_channel(self, pattern):
        for k in self.outputs:
            k.drop_channel(pattern)
    def delete_channel(self, pattern):
        self.drop_channel(pattern)
    def remove_channel(self, pattern):
        self.drop_channel(pattern)
            
    def add_gradient_channels(self, chanstr, new_name):
        for k in self.outputs:
            k.add_gradient_channel(chanstr, new_name)
        
    def num_timesteps(self):
        return [m.num_timesteps for m in self.outputs]

    def num_channels(self):
        if len(self.summary_stats) > 0:
            return [len([a for a,b in self.summary_stats.columns if b=='min'])]*self.noutputs
        else:
            return [m.num_channels for m in self.outputs]
    
    def elapsed_time(self):
        if len(self.summary_stats) > 0:
            return (self.summary_stats['Time']['max'].to_numpy() -
                    self.summary_stats['Time']['min'].to_numpy())
        else:
            return np.array( [m.elapsed_time for m in self.outputs] )
        
    def idxmins(self):
        return np.array( [m.idxmins for m in self.outputs] )
    
    def idxmaxs(self):
        return np.array( [m.idxmaxs for m in self.outputs] )
    
    def idxabs(self):
        return np.array( [m.idxabs for m in self.outputs] )
    
    def minima(self):
        if len(self.summary_stats) > 0:
            return np.array([self.summary_stats[a,b].to_list() for a,b in self.summary_stats.columns if b=='min']).T
        else:
            return np.array( [m.minima for m in self.outputs] )
        
    def maxima(self):
        if len(self.summary_stats) > 0:
            return np.array([self.summary_stats[a,b].to_list() for a,b in self.summary_stats.columns if b=='max']).T
        else:
            return np.array( [m.maxima for m in self.outputs] )
        
    def ranges(self):
        return (np.array(self.maxima()) - np.array(self.minima()))
    
    def variable(self):
        return np.array( [m.variable for m in self.outputs] )
    
    def constant(self):
        return np.array( [m.constant for m in self.outputs] )
    
    def sums(self):
        return np.array( [m.sums for m in self.outputs] )
    
    def sums_squared(self):
        return np.array( [m.sums_squared for m in self.outputs] )
    
    def sums_cubed(self):
        return np.array( [m.sums_cubed for m in self.outputs] )
    
    def sums_fourth(self):
        return np.array( [m.sums_fourth for m in self.outputs] )
    
    def second_moments(self):
        return np.array( [m.second_moments for m in self.outputs] )
    
    def third_moments(self):
        return np.array( [m.third_moments for m in self.outputs] )
    
    def fourth_moments(self):
        return np.array( [m.fourth_moments for m in self.outputs] )
    
    def means(self):
        if len(self.summary_stats) > 0:
            return np.array([self.summary_stats[a,b].to_list() for a,b in self.summary_stats.columns if b=='mean']).T
        else:
            return np.array( [m.means for m in self.outputs] )
        
    def medians(self):
        if len(self.summary_stats) > 0:
            return np.array([self.summary_stats[a,b].to_list() for a,b in self.summary_stats.columns if b=='median']).T
        else:
            return np.array( [m.medians for m in self.outputs] )
        
    def absmaxima(self):
        return np.array( [m.absmaxima for m in self.outputs] )
    
    def stddevs(self):
        if len(self.summary_stats) > 0:
            return np.array([self.summary_stats[a,b].to_list() for a,b in self.summary_stats.columns if b=='std']).T
        else:
            return np.array( [m.stddevs for m in self.outputs] )
        
    def skews(self):
        return np.array( [m.skews for m in self.outputs] )
    
    def kurtosis(self):
        return np.array( [m.kurtosis for m in self.outputs] )
    
    def integrated(self):
        if len(self.summary_stats) > 0:
            return np.array([self.summary_stats[a,b].to_list() for a,b in self.summary_stats.columns if b=='integrated']).T
        else:
            return np.array( [m.integrated for m in self.outputs] )
        
    def compute_energy(self, pwrchan):
        if len(self.summary_stats) > 0:
            return self.summary_stats[pwrchan]['integrated']
        else:
            return np.array( [m.compute_energy(pwrchan) for m in self.outputs] )
        
    def total_travel(self, chanstr):
        return np.array( [m.total_travel(chanstr) for m in self.outputs] )
        
    def time_averaging(self, time_window):
        for m in self.outputs:
            m.time_averaging(time_window)
    
    def time_binning(self, time_window):
        for m in self.outputs:
            m.time_binning(time_window)
    
    def psd(self, nfft=None):
        return [m.psd(nfft=nfft) for m in self.outputs]

